#!/usr/bin/env python
# coding=utf-8

# Copyright © 2012-2024 ButenkoMS. All rights reserved. Contacts: <gtalk@butenkoms.space>
# 
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
# 
#     http://www.apache.org/licenses/LICENSE-2.0
# 
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


__all__ = [
    'default_serializable_data_types',
    'known_types',
    'known_data_types',
    'known_container_types',
    'DataType',
    'data_type',
    'data_type_by_type',
    'ClassInfoFields',
    'ObjectInfoFields',
    'CanNotAdjustToSerializableError',
    'CanNotAdjustFromSerializableError',
    'RemoteObjectsManager',
]


from cengal.introspection.inspect import (
    entity_module_importable_str_and_owning_names_path,
    entity_by_name_module_importable_str_and_owning_names_path,
    filled_slot_names_with_values_gen,
    is_callable, is_async,
    is_setable_data_descriptor,
)
from cengal.code_flow_control.smart_values import ResultExistence
from cengal.code_flow_control.gc import DisableGC
from cengal.data_generation.id_generator import IDGenerator, GeneratorType
from enum import IntEnum
from collections.abc import MutableMapping, MutableSequence, MutableSet
from struct import pack, unpack
from inspect import getattr_static
from typing import Any, Dict, Optional, Callable, Set, Type, Tuple, List, FrozenSet


"""
Module Docstring
Docstrings: http://www.python.org/dev/peps/pep-0257/
"""

__author__ = "ButenkoMS <gtalk@butenkoms.space>"
__copyright__ = "Copyright © 2012-2024 ButenkoMS. All rights reserved. Contacts: <gtalk@butenkoms.space>"
__credits__ = ["ButenkoMS <gtalk@butenkoms.space>", ]
__license__ = "Apache License, Version 2.0"
__version__ = "4.4.0"
__maintainer__ = "ButenkoMS <gtalk@butenkoms.space>"
__email__ = "gtalk@butenkoms.space"
# __status__ = "Prototype"
__status__ = "Development"
# __status__ = "Production"


default_serializable_data_types: Set[Type] = {
    int, float, complex, str, bytes, bytearray, bool, type(None), list, tuple, set, frozenset, dict
}


known_types: Set[Type] = {
    int, float, complex, str, bytes, bytearray, bool, type(None), list, tuple, set, frozenset, dict,
    slice, 
}


known_data_types: Set[Type] = {
    int, float, str, bytes, bytearray, bool, type(None), slice,
}


known_container_types: Set[Type] = {
    complex, list, tuple, set, frozenset, dict,
}


class DataType(IntEnum):
    class_ = 0  # type is not in known_types and not in serializable_data_types
    int_ = 1
    float_ = 2
    complex_ = 3
    str_ = 4
    bytes_ = 5
    bytearray_ = 6
    bool_ = 7
    none_ = 8
    list_ = 9
    tuple_ = 10
    set_ = 11
    frozenset_ = 12
    dict_ = 13
    slice_ = 14
    unknown_serializable = 15  # type is in serializable_data_types but not in known_types


data_type: Dict[DataType, Type] = {
    0: object,
    1: int,
    2: float,
    3: complex,
    4: str,
    5: bytes,
    6: bytearray,
    7: bool,
    8: type(None),
    9: list,
    10: tuple,
    11: set,
    12: frozenset,
    13: dict,
    14: slice,
    15: None,
}


data_type_by_type: Dict[Type, DataType] = {
    object: 0,
    int: 1,
    float: 2,
    complex: 3,
    str: 4,
    bytes: 5,
    bytearray: 6,
    bool: 7,
    type(None): 8,
    list: 9,
    tuple: 10,
    set: 11,
    frozenset: 12,
    dict: 13,
    slice: 14,
    None: 15,
}


class ClassInfoFields(IntEnum):
    class_id = 0
    class_name = 1
    module_importable_str = 2
    owning_names_path = 3


class ObjectInfoFields(IntEnum):
    object_id = 0  # (type: int)
    type_id = 1  # (type: DataType)
    object_ = 2  # (Optional), (type: Any). Link to object itself if `(type(obj) in serializable_data_types)`. Not used otherwise.
    class_id = 3  # (Optional), (type: int). Used if `type_id == 0`
    clonable_slots = 4  # (Optional), (type: Tuple[Tuple[str, Any]]). Used if `type_id == 0`. Holds ID's (object_id) of slots objects.
    clonable_dict_items = 5  # (Optional), (type: Tuple[Tuple[str, Any]]). Used if `type_id == 0`. Holds ID's (object_id) of value objects.
    # contained_items = 6  # (Optional), (type: Union[Tuple, List, Set, FrozenSet, Dict]). Used if `type_id in {9, 10, 11, 12, 13, 14}`. Holds ID's (object_id) of contained items (for bothe keys and values in the case of Dict).
    contained_mapping = 6  # (Optional), (type: Union[Tuple, List, Set, FrozenSet, Dict]). Used if `type_id in {9, 10, 11, 12, 13, 14}`. Holds ID's (object_id) of contained items (for bothe keys and values in the case of Dict).
    contained_sequence = 7  # (Optional), (type: Union[Tuple, List, Set, FrozenSet, Dict]). Used if `type_id in {9, 10, 11, 12, 13, 14}`. Holds ID's (object_id) of contained items (for bothe keys and values in the case of Dict).
    contained_set = 8  # (Optional), (type: Union[Tuple, List, Set, FrozenSet, Dict]). Used if `type_id in {9, 10, 11, 12, 13, 14}`. Holds ID's (object_id) of contained items (for bothe keys and values in the case of Dict).


class CanNotAdjustToSerializableError(Exception):
    pass


class CanNotAdjustFromSerializableError(Exception):
    pass


class RemoteObjectsManager:
    __slots__ = (
        'classes_id_gen',
        'objects_db',
        'objects_id_gen',
        'serializable_data_types',
        'serializable_any',
        'serializable_int',
        'serializable_float',
        'serializable_complex',
        'serializable_str',
        'serializable_bytes',
        'serializable_bytearray',
        'serializable_bool',
        'serializable_none',
        'serializable_list',
        'serializable_tuple',
        'serializable_set',
        'serializable_frozenset',
        'serializable_dict',
        'serializable_slice',
        'known_classes',
        'known_classes_by_id',
        'known_classes_info',
        'known_classes_info_by_id',
        'on_new_class_handler',
        'on_new_obj_info_handler',
        'objects_ids',
        'objects_ids_by_id',
    )

    def __init__(self, 
                 on_new_class_handler: Optional[Callable] = None,
                 on_new_obj_info_handler: Optional[Callable] = None,
                 # classess_db: Optional[Dict[int, Type]] = None, 
                 objects_db: Optional[Dict[int, Any]] = None, 
                 classes_id_gen: Optional[Callable] = None,
                 objects_id_gen: Optional[Callable] = None,
                 serializable_data_types: Optional[Set[Type]] = None,
                 ) -> None:
        # self.classess_db: Dict[int, Type] = dict() if classess_db is None else classess_db
        self.classes_id_gen: IDGenerator = IDGenerator() if classes_id_gen is None else classes_id_gen
        self.objects_db: Dict[int, Any] = dict() if objects_db is None else objects_db
        self.objects_id_gen: IDGenerator = IDGenerator() if objects_id_gen is None else objects_id_gen
        self.serializable_data_types: Set[Type] = default_serializable_data_types if serializable_data_types is None else serializable_data_types
        self.serializable_any: bool = not self.serializable_data_types
        self.serializable_int: bool = (int in self.serializable_data_types) or self.serializable_any
        self.serializable_float: bool = (float in self.serializable_data_types) or self.serializable_any
        self.serializable_complex: bool = (complex in self.serializable_data_types) or self.serializable_any
        self.serializable_str: bool = (str in self.serializable_data_types) or self.serializable_any
        self.serializable_bytes: bool = (bytes in self.serializable_data_types) or self.serializable_any
        self.serializable_bytearray: bool = (bytearray in self.serializable_data_types) or self.serializable_any
        self.serializable_bool: bool = (bool in self.serializable_data_types) or self.serializable_any
        self.serializable_none: bool = (type(None) in self.serializable_data_types) or self.serializable_any
        self.serializable_list: bool = (list in self.serializable_data_types) or self.serializable_any
        self.serializable_tuple: bool = (tuple in self.serializable_data_types) or self.serializable_any
        self.serializable_set: bool = (set in self.serializable_data_types) or self.serializable_any
        self.serializable_frozenset: bool = (frozenset in self.serializable_data_types) or self.serializable_any
        self.serializable_dict: bool = (dict in self.serializable_data_types) or self.serializable_any
        self.serializable_slice: bool = (slice in self.serializable_data_types) or self.serializable_any
        self.known_classes: Dict[Type, int] = dict()
        self.known_classes_by_id: Dict[int, Type] = dict()
        self.known_classes_info: Dict[Tuple, int] = dict()
        self.known_classes_info_by_id: Dict[int, Tuple] = dict()
        self.on_new_class_handler: Optional[Callable] = on_new_class_handler
        self.on_new_obj_info_handler: Optional[Callable] = on_new_obj_info_handler
        self.objects_ids: Dict[int, int] = dict()  # (Key: object_id, Value: id(obj))
        self.objects_ids_by_id: Dict[int, int] = dict()  # (Key: id(obj), Value: object_id)

    def del_object_by_id(self, id_: int) -> None:
        if id_ in self.objects_ids_by_id:
            object_id = self.objects_ids_by_id[id_]
            self.objects_ids_by_id.pop(id_, None)
            self.objects_ids.pop(object_id, None)
            self.objects_db.pop(object_id, None)
    
    dobi = del_object_by_id
    
    def del_object_by_object_id(self, object_id: int) -> None:
        if object_id in self.objects_ids:
            id_ = self.objects_ids[object_id]
            self.objects_ids_by_id.pop(id_, None)
            self.objects_ids.pop(object_id, None)
            self.objects_db.pop(object_id, None)
    
    doboi = del_object_by_object_id
    
    def adjust_to_serializable(self, type_id: DataType, obj: Any) -> Any:
        if 1 == type_id:
            if self.serializable_int:
                return obj
            elif self.serializable_bytes:
                return obj.to_bytes(8, 'little')
            elif self.serializable_bytearray:
                return bytearray(obj.to_bytes(8, 'little'))
            elif self.serializable_str:
                return str(obj)
            else:
                pass
        elif 2 == type_id:
            if self.serializable_float:
                return obj
            elif self.serializable_bytes:
                return pack('=d', obj)
            elif self.serializable_bytearray:
                return bytearray(pack('=d', obj))
            elif self.serializable_str:
                return str(obj)
            else:
                pass
        elif 3 == type_id:
            if self.serializable_complex:
                return obj
            elif self.serializable_bytes:
                return pack('=dd', obj.real, obj.imag)
            elif self.serializable_bytearray:
                return bytearray(pack('=dd', obj.real, obj.imag))
            elif self.serializable_str:
                return str(obj)
            elif self.serializable_float:
                if self.serializable_tuple:
                    return (obj.real, obj.imag)
                elif self.serializable_list:
                    return [obj.real, obj.imag]
                else:
                    pass
            else:
                pass
        elif 4 == type_id:
            if self.serializable_str:
                return obj
            elif self.serializable_bytes:
                return obj.encode('utf-8')
            elif self.serializable_bytearray:
                return bytearray(obj.encode('utf-8'))
            elif self.serializable_int:
                if self.serializable_tuple:
                    return tuple(ord(c) for c in obj)
                elif self.serializable_list:
                    return [ord(c) for c in obj]
                else:
                    pass
            elif self.serializable_float:
                if self.serializable_tuple:
                    return tuple(float(ord(c)) for c in obj)
                elif self.serializable_list:
                    return [float(ord(c)) for c in obj]
                else:
                    pass
            else:
                pass
        elif 5 == type_id:
            if self.serializable_bytes:
                return obj
            elif self.serializable_bytearray:
                return bytearray(obj)
            elif self.serializable_str:
                return obj.hex()
            elif self.serializable_tuple:
                if self.serializable_int:
                    return  tuple(int(c) for c in obj)
                if self.serializable_float:
                    return  tuple(float(int(c)) for c in obj)
                else:
                    pass
            elif self.serializable_list:
                if self.serializable_int:
                    return  [int(c) for c in obj]
                if self.serializable_float:
                    return  [float(int(c)) for c in obj]
                else:
                    pass
            else:
                pass
        elif 6 == type_id:
            if self.serializable_bytearray:
                return obj
            elif self.serializable_bytes:
                return bytes(obj)
            elif self.serializable_str:
                return obj.hex()
            elif self.serializable_tuple:
                if self.serializable_int:
                    return  tuple(int(c) for c in obj)
                if self.serializable_float:
                    return  tuple(float(int(c)) for c in obj)
                else:
                    pass
            elif self.serializable_list:
                if self.serializable_int:
                    return  [int(c) for c in obj]
                if self.serializable_float:
                    return  [float(int(c)) for c in obj]
                else:
                    pass
            else:
                pass
        elif 7 == type_id:
            if self.serializable_bool:
                return obj
            elif self.serializable_int:
                return int(obj)
            elif self.serializable_bytes:
                return obj.to_bytes(1, 'little')
            elif self.serializable_bytearray:
                return bytearray(obj.to_bytes(1, 'little'))
            elif self.serializable_str:
                return str(obj)
            elif self.serializable_float:
                return float(int(obj))
            else:
                pass
        elif 8 == type_id:
            if self.serializable_none:
                return obj
            elif self.serializable_bool:
                return False
            elif self.serializable_int:
                return 0
            elif self.serializable_bytes:
                return b'\x00'
            elif self.serializable_bytearray:
                return bytearray(b'\x00')
            elif self.serializable_str:
                return ''
            elif self.serializable_float:
                return 0.0
            else:
                pass
        elif 9 == type_id:
            if self.serializable_list:
                return obj
            elif self.serializable_tuple:
                return tuple(obj)
            elif self.serializable_dict:
                return dict({index: item for index, item in enumerate(obj)})
            else:
                pass
        elif 10 == type_id:
            if self.serializable_tuple:
                return obj
            elif self.serializable_list:
                return list(obj)
            elif self.serializable_dict:
                return dict({index: item for index, item in enumerate(obj)})
            else:
                pass
        elif 11 == type_id:
            if self.serializable_set:
                return obj
            elif self.serializable_frozenset:
                return frozenset(obj)
            elif self.serializable_tuple:
                return tuple(obj)
            elif self.serializable_list:
                return list(obj)
            elif self.serializable_dict:
                return dict({k: None for k in obj})
            else:
                pass
        elif 12 == type_id:
            if self.serializable_frozenset:
                return obj
            elif self.serializable_set:
                return set(obj)
            elif self.serializable_tuple:
                return tuple(obj)
            elif self.serializable_list:
                return list(obj)
            elif self.serializable_dict:
                return dict({k: None for k in obj})
            else:
                pass
        elif 13 == type_id:
            if self.serializable_dict:
                return obj
            elif self.serializable_tuple:
                return tuple(obj.items())
            elif self.serializable_list:
                return list(list(pair) for pair in obj.items())
            else:
                pass
        elif 14 == type_id:
            if self.serializable_slice:
                return obj
            elif self.serializable_tuple:
                return (obj.start, obj.stop, obj.step)
            elif self.serializable_list:
                return [obj.start, obj.stop, obj.step]
            elif self.serializable_dict:
                return {0: obj.start, 1: obj.stop, 2: obj.step}
            else:
                pass
        else:
            raise RuntimeError('Unknown type_id')
        
        raise CanNotAdjustToSerializableError(f'Can not adjust to serializable. Type: {type_id}, obj: {obj}')

    ats = adjust_to_serializable

    def adjust_from_serializable(self, type_id: DataType, obj: Any) -> Any:
        if 1 == type_id:
            if self.serializable_int:
                return obj
            elif self.serializable_bytes:
                return int.from_bytes(obj, 'little')
            elif self.serializable_bytearray:
                return int.from_bytes(bytes(obj), 'little')
            elif self.serializable_str:
                return int(obj)
            else:
                pass
        elif 2 == type_id:
            if self.serializable_float:
                return obj
            elif self.serializable_bytes:
                return unpack('=d', obj)[0]
            elif self.serializable_bytearray:
                return unpack('=d', bytes(obj))[0]
            elif self.serializable_str:
                return float(obj)
            else:
                pass
        elif 3 == type_id:
            if self.serializable_complex:
                return obj
            elif self.serializable_bytes:
                return complex(*unpack('=dd', obj))
            elif self.serializable_bytearray:
                return complex(*unpack('=dd', bytes(obj)))
            elif self.serializable_str:
                return complex(*obj)
            elif self.serializable_float:
                if self.serializable_tuple:
                    return complex(*obj)
                elif self.serializable_list:
                    return complex(*obj)
                else:
                    pass
            else:
                pass
        elif 4 == type_id:
            if self.serializable_str:
                return obj
            elif self.serializable_bytes:
                return obj.decode('utf-8')
            elif self.serializable_bytearray:
                return bytes(obj).decode('utf-8')
            elif self.serializable_int:
                if self.serializable_tuple:
                    return ''.join(chr(c) for c in obj)
                elif self.serializable_list:
                    return ''.join(chr(c) for c in obj)
                else:
                    pass
            elif self.serializable_float:
                if self.serializable_tuple:
                    return ''.join(chr(int(round(c))) for c in obj)
                elif self.serializable_list:
                    return ''.join(chr(int(round(c))) for c in obj)
                else:
                    pass
            else:
                pass
        elif 5 == type_id:
            if self.serializable_bytes:
                return obj
            elif self.serializable_bytearray:
                return bytes(obj)
            elif self.serializable_str:
                return bytes.fromhex(obj)
            elif self.serializable_tuple:
                if self.serializable_int:
                    return  b''.join(item.to_bytes(1, 'little') for item in obj)
                if self.serializable_float:
                    return  b''.join((int(round(item))).to_bytes(1, 'little') for item in obj)
                else:
                    pass
            elif self.serializable_list:
                if self.serializable_int:
                    return  b''.join(item.to_bytes(1, 'little') for item in obj)
                if self.serializable_float:
                    return  b''.join((int(round(item))).to_bytes(1, 'little') for item in obj)
                else:
                    pass
            else:
                pass
        elif 6 == type_id:
            if self.serializable_bytearray:
                return obj
            elif self.serializable_bytes:
                return bytearray(obj)
            elif self.serializable_str:
                return bytearray(bytes.fromhex(obj))
            elif self.serializable_tuple:
                if self.serializable_int:
                    return  bytearray(b''.join(item.to_bytes(1, 'little') for item in obj))
                if self.serializable_float:
                    return  bytearray(b''.join((int(round(item))).to_bytes(1, 'little') for item in obj))
                else:
                    pass
            elif self.serializable_list:
                if self.serializable_int:
                    return  bytearray(b''.join(item.to_bytes(1, 'little') for item in obj))
                if self.serializable_float:
                    return  bytearray(b''.join((int(round(item))).to_bytes(1, 'little') for item in obj))
                else:
                    pass
            else:
                pass
        elif 7 == type_id:
            if self.serializable_bool:
                return obj
            elif self.serializable_int:
                return bool(obj)
            elif self.serializable_bytes:
                return bool(int.from_bytes(obj, 'little'))
            elif self.serializable_bytearray:
                return bool(int.from_bytes(bytes(obj), 'little'))
            elif self.serializable_str:
                return bool(obj)
            elif self.serializable_float:
                return bool(int(round(obj)))
            else:
                pass
        elif 8 == type_id:
            return None
        elif 9 == type_id:
            if self.serializable_list:
                return obj
            elif self.serializable_tuple:
                return list(obj)
            elif self.serializable_dict:
                return [value for key, value in sorted(obj.items(), key=lambda x: x[0])]
            else:
                pass
        elif 10 == type_id:
            if self.serializable_tuple:
                return obj
            elif self.serializable_list:
                return tuple(obj)
            elif self.serializable_dict:
                return tuple(value for key, value in sorted(obj.items(), key=lambda x: x[0]))
            else:
                pass
        elif 11 == type_id:
            if self.serializable_set:
                return obj
            elif self.serializable_frozenset:
                return set(obj)
            elif self.serializable_tuple:
                return set(obj)
            elif self.serializable_list:
                return set(obj)
            elif self.serializable_dict:
                return set(obj.keys())
            else:
                pass
        elif 12 == type_id:
            if self.serializable_frozenset:
                return obj
            elif self.serializable_set:
                return frozenset(obj)
            elif self.serializable_tuple:
                return frozenset(obj)
            elif self.serializable_list:
                return frozenset(obj)
            elif self.serializable_dict:
                return frozenset(obj.keys())
            else:
                pass
        elif 13 == type_id:
            if self.serializable_dict:
                return obj
            elif self.serializable_tuple:
                return dict(obj)
            elif self.serializable_list:
                return dict(obj)
            else:
                pass
        elif 14 == type_id:
            if self.serializable_slice:
                return obj
            elif self.serializable_tuple:
                return slice(*obj)
            elif self.serializable_list:
                return slice(*obj)
            elif self.serializable_dict:
                return slice(obj[0], obj[1], obj[2])
            else:
                pass
        else:
            raise RuntimeError('Unknown type_id')
        
        raise CanNotAdjustFromSerializableError(f'Can not adjust from serializable. Type: {type_id}, obj: {obj}')
    
    afs = adjust_from_serializable
    
    # def is_replicatable_object_attribute(self, attribute: Any) -> bool:
    #     if is_setable_data_descriptor(attribute):
    #         data = attribute.__get__(None, None)
    #     elif is_callable(attribute):
    #         return False
    #     else:
    #         return True

    def serialize_container(self, type_id: DataType, obj: Any) -> Any:
        if 9 == type_id:
            new_obj = list()
            for item in obj:
                exists, value = self.serialize_impl(item)
                if exists:
                    new_obj.append(value)
        elif 10 == type_id:
            new_obj = list()
            for item in obj:
                exists, value = self.serialize_impl(item)
                if exists:
                    new_obj.append(value)
            
            new_obj = tuple(new_obj)
        elif 11 == type_id:
            new_obj = set()
            for item in obj:
                exists, value = self.serialize_impl(item)
                if exists:
                    new_obj.add(value)
        elif 12 == type_id:
            new_obj = set()
            for item in obj:
                exists, value = self.serialize_impl(item)
                if exists:
                    new_obj.add(value)
                
            new_obj = frozenset(new_obj)
        elif 13 == type_id:
            new_obj = dict()
            for key, value in obj.items():
                key_exists, key_value = self.serialize_impl(key)
                value_exists, value_value = self.serialize_impl(value)
                if key_exists and value_exists:
                    new_obj[key_value] = value_value
        
        return new_obj
    
    sc = serialize_container

    def serialize_impl(self, obj: Any, ignore_empty_classes: bool = False) -> Tuple[bool, int]:
        result_exists: bool = True
        id_: int = id(obj)
        obj_type = type(obj)
        if (int != obj_type) and (id_ in self.objects_ids_by_id):
            new_object: bool = False
            object_id: int = self.objects_ids_by_id[id_]
        else:
            # int object must always produce new object_id because first 256 ints are persistent across Python sessions and
            # this can cause issues within users of current module. For example within `cengal/hardware/memory/shared_memory`
            # which changes int values inline instead of producing new objects.
            new_object = True
            object_id = self.objects_id_gen()
            self.objects_ids[object_id] = id_
            self.objects_db[object_id] = obj
        
        if not new_object:
            return ResultExistence[int](result_exists, object_id)
        
        if  self.serializable_any or (obj_type in self.serializable_data_types):
            serializable: bool = True
            type_id: int = data_type_by_type.get(obj_type, 15)
        else:
            serializable = False
            type_id = data_type_by_type.get(obj_type, 0)
        
        if obj_type in known_container_types:
            known_container: bool = True
        else:
            known_container = False
        
        class_info: Dict[int, Any] = None
        known_data: bool = None
        class_id: int = None
        is_new_class: bool = None
        class_name: str = None
        new_obj_slots: List[Tuple[str, Any]] = None
        new_obj_dict: Dict[str, Any] = None
        obj_mapping: Dict = None
        obj_sequence: List = None
        obj_set: Set = None
        if serializable:
            if known_container:
                object_info = (
                    self.ats(1, object_id),
                    self.ats(1, type_id),
                    self.sc(type_id, obj),
                )
            else:
                object_info = (
                    self.ats(1, object_id),
                    self.ats(1, type_id),
                    obj,
                )
        else:
            if obj_type in known_data_types:
                known_data = True
            else:
                known_data = False
            
            if 0 == type_id:
                new_obj_slots = list()
                for slot_name, slot_value in filled_slot_names_with_values_gen(obj):
                    # if not hasattr(slot_value, '__set__'):
                    #     # if not setable descriptor
                    #     continue

                    if self.serializable_str:
                        adjusted_slot_name = slot_name
                    else:
                        adjusted_slot_name = self.ats(4, slot_name)
                    
                    item_exists, item_value = self.serialize_impl(slot_value)
                    if item_exists:
                        if self.serializable_tuple:
                            new_obj_slots.append((adjusted_slot_name, item_value))
                        elif self.serializable_list:
                            new_obj_slots.append([adjusted_slot_name, item_value])
                        else:
                            new_obj_slots.append(self.ats(10, (adjusted_slot_name, item_value)))
                
                if new_obj_slots:
                    if self.serializable_list:
                        pass
                    elif self.serializable_tuple:
                        new_obj_slots = tuple(new_obj_slots)
                    else:
                        new_obj_slots = self.ats(9, new_obj_slots)
                
                new_obj_dict = dict()
                if hasattr(obj, '__dict__'):
                    for key, value in obj.__dict__.items():
                        # raw_value = getattr_static(obj, key)
                        # if hasattr(raw_value, '__get__') and (not hasattr(raw_value, '__set__')):
                        #     # if not setable descriptor
                        #     continue

                        if self.serializable_str:
                            adjusted_key = key
                        else:
                            adjusted_key = self.ats(4, key)

                        value_exists, value_value = self.serialize_impl(value)
                        if value_exists:
                            new_obj_dict[adjusted_key] = value_value
                        else:
                            continue
                
                    if new_obj_dict:
                        if self.serializable_dict:
                            pass
                        else:
                            new_obj_dict = self.ats(13, new_obj_dict)
                
                obj_mapping = dict()
                obj_sequence = list()
                obj_set = set()
                if isinstance(obj, MutableMapping):
                    for key, value in obj.items():
                        key_exists, key_value = self.serialize_impl(value)
                        if not key_exists:
                            continue

                        value_exists, value_value = self.serialize_impl(value)
                        if value_exists:
                            obj_mapping[key_value] = value_value
                        else:
                            continue
                    
                    if obj_mapping:
                        if self.serializable_dict:
                            pass
                        else:
                            obj_mapping = self.ats(13, obj_mapping)
                elif isinstance(obj, MutableSequence):
                    for item in obj:
                        item_exists, item_value = self.serialize_impl(item)
                        if item_exists:
                            obj_sequence.append(item_value)
                        else:
                            continue
                    
                    if obj_sequence:
                        if self.serializable_list:
                            pass
                        else:
                            obj_sequence = self.ats(9, obj_sequence)
                elif isinstance(obj, MutableSet):
                    for item in obj:
                        item_exists, item_value = self.serialize_impl(item)
                        if item_exists:
                            obj_set.add(item_value)
                        else:
                            continue
                    
                    if obj_set:
                        if self.serializable_set:
                            pass
                        else:
                            obj_set = self.ats(11, obj_set)
                else:
                    pass

                if ignore_empty_classes:
                    result_exists = new_obj_slots or new_obj_dict or obj_mapping or obj_sequence or obj_set
                
                if obj_type in self.known_classes:
                    is_new_class = False
                    class_id = self.known_classes[obj_type]
                else:
                    is_new_class = True
                    class_name = obj_type.__name__
                    class_id = self.classes_id_gen()
                    self.known_classes[obj_type] = class_id
                    self.known_classes_by_id[class_id] = obj_type
                    module_importable_str, owning_names_path = entity_module_importable_str_and_owning_names_path(obj)
                    adjusted_owning_names_path: List[str] = [self.ats(4, name) for name in owning_names_path]
                    module_importable_str_and_owning_names_path = (class_name, module_importable_str, tuple(owning_names_path))
                    self.known_classes_info[module_importable_str_and_owning_names_path] = class_id
                    self.known_classes_info_by_id[class_id] = module_importable_str_and_owning_names_path

                    if result_exists:
                        class_info = (
                            self.ats(1, class_id),
                            self.ats(4, class_name),
                            self.ats(4, module_importable_str),
                            self.ats(9, adjusted_owning_names_path),
                        )

                        if self.serializable_tuple:
                            adjusted_class_info = class_info
                        else:
                            adjusted_class_info = self.ats(10, class_info)
                        
                        if self.on_new_class_handler:
                            self.on_new_class_handler(
                                adjusted_class_info,
                                obj_type,
                                class_id,
                                class_name,
                                module_importable_str,
                                owning_names_path,
                                module_importable_str_and_owning_names_path
                            )

                # will be later serialized much faster than without `if else None`
                object_info = (
                    self.ats(1, object_id),
                    self.ats(1, type_id),
                    self.ats(8, None),
                    self.ats(1, class_id),
                    new_obj_slots if new_obj_slots else self.ats(9, list()),
                    new_obj_dict if new_obj_dict else self.ats(13, dict()),
                    obj_mapping if obj_mapping else self.ats(13, dict()),
                    obj_sequence if obj_sequence else self.ats(9, list()),
                    obj_set if obj_set else self.ats(11, set()),
                )
            elif known_data:
                object_info = (
                    self.ats(1, object_id),
                    self.ats(1, type_id),
                    self.ats(type_id, obj),
                )
            elif known_container:
                object_info = (
                    self.ats(1, object_id),
                    self.ats(1, type_id),
                    self.ats(type_id, self.sc(type_id, obj)),
                )
            else:
                raise RuntimeError('Unknown type_id')

        if self.serializable_tuple:
            adjusted_object_info = object_info
        else:
            adjusted_object_info = self.ats(10, object_info)
        
        if result_exists:
            if self.on_new_obj_info_handler:
                self.on_new_obj_info_handler(
                    adjusted_object_info,
                    obj,
                    object_id,
                    type_id,
                    serializable,
                    known_container,
                    known_data,
                    class_id,
                    is_new_class,
                    new_obj_slots,
                    new_obj_dict,
                    obj_mapping,
                    obj_sequence,
                    obj_set,
                )
        
        return result_exists, object_id
    
    si = serialize_impl

    def serialize(self, obj: Any, ignore_empty_classes: bool = False) -> Tuple[bool, int]:
        with DisableGC():
            return self.serialize_impl(obj, ignore_empty_classes)
    
    s = serialize
    
    def deserialize_class_impl(self, class_info: Tuple) -> Tuple[bool, int, Type]:
        if self.serializable_tuple:
            pass
        else:
            class_info = self.afs(10, class_info)

        class_id: int = self.afs(1, class_info[0])
        if class_id in self.known_classes_by_id:
            return ResultExistence[Tuple[int, Type]](True, (class_id, self.known_classes_by_id[class_id]))
        else:
            class_name: str = self.afs(4, class_info[1])
            module_importable_str: str = self.afs(4, class_info[2])
            owning_names_path: List[str] = self.afs(9, class_info[3])
            adjusted_owning_names_path: List[str] = [self.afs(4, name) for name in owning_names_path]
            obj_class: Type = entity_by_name_module_importable_str_and_owning_names_path(class_name, module_importable_str, adjusted_owning_names_path)
            self.known_classes_by_id[class_id] = obj_class
            self.known_classes[obj_class] = class_id
            class_info_tuple: Tuple = (class_name, module_importable_str, tuple(owning_names_path)) 
            self.known_classes_info[class_info_tuple] = class_id
            self.known_classes_info_by_id[class_id] = class_info_tuple
            return True, class_id, obj_class
    
    dcli = deserialize_class_impl
    
    def deserialize_class(self, class_info: Tuple) -> Tuple[bool, int, Type]:
        with DisableGC():
            return self.deserialize_class_impl(class_info)
    
    dcl = deserialize_class

    def deserialize_container_impl(self, type_id: DataType, obj: Any) -> Any:
        new_obj = obj
        if 9 == type_id:
            new_obj = list()
            for item_id in obj:
                new_obj.append(self.objects_db[self.afs(1, item_id)])
        elif 10 == type_id:
            new_obj = list()
            for item_id in obj:
                new_obj.append(self.objects_db[self.afs(1, item_id)])
            
            new_obj = tuple(new_obj)
        elif 11 == type_id:
            new_obj = set()
            for item_id in obj:
                new_obj.add(self.objects_db[self.afs(1, item_id)])
        elif 12 == type_id:
            new_obj = set()
            for item_id in obj:
                new_obj.add(self.objects_db[self.afs(1, item_id)])
                
            new_obj = frozenset(new_obj)
        elif 13 == type_id:
            new_obj = dict()
            for key_id, value_id in obj.items():
                new_obj[self.objects_db[int(self.afs(1, key_id))]] = self.objects_db[self.afs(1, value_id)]
        
        return new_obj
    
    dcoi = deserialize_container_impl

    def deserialize_container(self, type_id: DataType, obj: Any) -> Any:
        with DisableGC():
            return self.deserialize_container_impl(type_id, obj)
    
    dco = deserialize_container

    def deserialize_obj_impl(self, obj_info: Tuple) -> Tuple[bool, int, Any]:
        if self.serializable_tuple:
            pass
        else:
            obj_info = self.afs(10, obj_info)
        
        object_id: int = self.afs(1, obj_info[0])
        if object_id in self.objects_db:
            return ResultExistence[Tuple[int, Any]](True, (object_id, self.objects_db[object_id]))
        
        type_id: int = self.afs(1, obj_info[1])
        
        obj_type = data_type[type_id] 
        serializable: bool = self.serializable_any or (obj_type in self.serializable_data_types)
        known_container: bool = obj_type in known_container_types
        known_data: bool = None
        
        if serializable:
            if known_container:
                obj = self.dcoi(type_id, obj_info[2])
            else:
                obj = obj_info[2]
        else:
            known_data = obj_type in known_data_types
            
            if 0 == type_id:
                class_id: int = self.afs(1, obj_info[3])
                obj_class: Type = self.known_classes_by_id[class_id]
                obj: Any = obj_class.__new__(obj_class)

                if obj_info[4]:
                    clonable_slots = self.afs(9, obj_info[4])
                    for slot_name, slot_id in clonable_slots:
                        slot_name = self.afs(4, slot_name)
                        slot_id = self.afs(1, slot_id)
                        child_obj = self.objects_db[slot_id]
                        setattr(obj, slot_name, child_obj)
                
                if obj_info[5]:
                    clonable_dict_items = self.afs(13, obj_info[5])
                    for key, value_id in clonable_dict_items.items():
                        key = self.afs(4, key)
                        value_id = self.afs(1, value_id)
                        child_obj = self.objects_db[value_id]
                        setattr(obj, key, child_obj)
                
                if obj_info[6]:
                    contained_mapping = self.afs(13, obj_info[6])
                    for key_id, value_id in contained_mapping.items():
                        key_id = self.afs(1, key_id)
                        value_id = self.afs(1, value_id)
                        child_key = self.objects_db[int(key_id)]
                        child_value = self.objects_db[value_id]
                        obj[child_key] = child_value
                
                if obj_info[7]:
                    contained_sequence = self.afs(9, obj_info[7])
                    for item_id in contained_sequence:
                        item_id = self.afs(1, item_id)
                        child_item = self.objects_db[item_id]
                        obj.append(child_item)
                
                if obj_info[8]:
                    contained_set = self.afs(11, obj_info[8])
                    for item_id in contained_set:
                        item_id = self.afs(1, item_id)
                        child_item = self.objects_db[item_id]
                        obj.add(child_item)
            elif known_data:
                obj = self.afs(type_id, obj_info[2])
            elif known_container:
                obj = self.dcoi(type_id, self.afs(type_id, obj_info[2]))
            else:
                raise RuntimeError('Unknown type_id')
        
        self.objects_db[object_id] = obj
        id_: int = id(obj)
        self.objects_ids[object_id] = id_
        self.objects_ids_by_id[id_] = object_id
        return True, object_id, obj
    
    doi = deserialize_obj_impl

    def deserialize_obj(self, obj_info: Tuple) -> Tuple[bool, int, Any]:
        with DisableGC():
            return self.deserialize_obj_impl(obj_info)
    
    do = deserialize_obj



#!/usr/bin/env python
# coding=utf-8

# Copyright © 2012-2024 ButenkoMS. All rights reserved. Contacts: <gtalk@butenkoms.space>
# 
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
# 
#     http://www.apache.org/licenses/LICENSE-2.0
# 
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


__all__ = [
    'CoroApp', 
    'bind_coro_explicit', 
    'bind_coro_implicit', 
    'bind_coro', 
    'abind_coro_explicit', 
    'abind_coro_implicit', 
    'abind_coro', 
    'asyncio_bind_coro_explicit', 
    'asyncio_bind_coro_implicit', 
    'asyncio_bind_coro', 
    'bind_asyncio_coro', 
    'bind_running_coro', 
    'bind_running_asyncio_coro', 
    'wx_exec_in_coro', 
    'bind_to', 
    'bind', 
    'event_handler_implicit', 
    'event_handler', 
    'event_handler_explicit', 
    'asyncio_event_handler', 
    'aevent_handler', 
    'blocking_event_handler_implicit', 
    'blocking_event_handler_explicit', 
]


"""
Module Docstring
Docstrings: http://www.python.org/dev/peps/pep-0257/
"""

__author__ = "ButenkoMS <gtalk@butenkoms.space>"
__copyright__ = "Copyright © 2012-2024 ButenkoMS. All rights reserved. Contacts: <gtalk@butenkoms.space>"
__credits__ = ["ButenkoMS <gtalk@butenkoms.space>", ]
__license__ = "Apache License, Version 2.0"
__version__ = "4.4.0"
__maintainer__ = "ButenkoMS <gtalk@butenkoms.space>"
__email__ = "gtalk@butenkoms.space"
# __status__ = "Prototype"
__status__ = "Development"
# __status__ = "Production"


from cengal.parallel_execution.coroutines.coro_scheduler import CoroScheduler, CoroSchedulerType, Interface, CoroID, Coro, AnyWorker, current_interface, current_coro_scheduler, cs_coro, cs_acoro
from cengal.parallel_execution.coroutines.coro_standard_services.loop_yield import gly, CoroPriority, gly_patched, agly_patched
from cengal.parallel_execution.coroutines.coro_standard_services.run_coro import RunCoro
from cengal.parallel_execution.coroutines.coro_standard_services.put_coro import PutCoro
from cengal.parallel_execution.coroutines.coro_standard_services.asyncio_loop import AsyncioLoopRequest
from cengal.parallel_execution.coroutines.coro_standard_services.async_event_bus import AsyncEventBusRequest, try_send_async_event
from cengal.parallel_execution.coroutines.coro_standard_services.simple_yield import Yield
from cengal.parallel_execution.coroutines.coro_standard_services.shutdown_loop import ShutdownLoop
from cengal.parallel_execution.coroutines.coro_standard_services.instance import InstanceRequest
from cengal.parallel_execution.coroutines.coro_tools.await_coro import await_task_prim
from cengal.parallel_execution.coroutines.coro_tools.run_in_loop import run_in_loop
from cengal.parallel_execution.coroutines.coro_tools.coro_flow_control import graceful_coro_destroyer
from cengal.code_flow_control.smart_values import ValueHolder
from cengal.data_generation.id_generator import IDGenerator
from cengal.time_management.sleep_tools import try_sleep, get_usable_min_sleep_interval
from cengal.math.numbers import RationalNumber

from math import ceil
from inspect import signature, Signature, ismethod
from contextlib import contextmanager, asynccontextmanager
from functools import wraps, update_wrapper, partial
from asyncio import Future, Task, create_task
from typing import Optional, Any, Callable, Hashable, Set, Union, Awaitable, Coroutine


from wx import App, Window, Frame, Dialog, Timer, GetTopLevelParent, EVT_TIMER, EVT_CLOSE, ID_ANY


class BindedEntityIsAboutToBeDestroyed(Exception):
    pass


def sec_to_ms(sec: RationalNumber) -> int:
    return int(ceil(sec * 1000))


class CoroApp(App):
    def __init__(self, default_priority: CoroPriority = CoroPriority.normal):
        self._cs: CoroSchedulerType = current_coro_scheduler()
        self._cs.on_idle_handlers.add(self._on_system_loop_idle)
        self._cs.idle_managers_num += 1
        self._default_priority: CoroPriority = default_priority
        self._ly = gly(self._default_priority)
        self._binded_coros: Set = set()
        self._idle_for: Optional[int] = None  # in milliseconds
        super(CoroApp, self).__init__()

    def OnInit(self):
        super().OnInit()
        self._system_loop_timer = Timer(self)
        self.Bind(EVT_TIMER, self._on_system_loop_timer, self._system_loop_timer)
        self._system_loop_timer.StartOnce(1) # run every 1 millisecond
        return True

    def OnExit(self):
        super().OnExit()
        i: Interface = current_interface()
        i.log.info('Application is exiting.')
        i.log.info('Destroying all binded coros...')
        for furure_or_coro_id in self._binded_coros:
            if isinstance(furure_or_coro_id, Future):
                furure_or_coro_id.cancel()
            else:
                graceful_coro_destroyer(i, 0.1, furure_or_coro_id, BindedEntityIsAboutToBeDestroyed)
        
        i.log.info('All binded coroutines destroyed.')
        self._cs.idle_managers_num -= 1
        self._cs.on_idle_handlers.discard(self._on_system_loop_idle)
        i.log.info('Application is exiting... Done.')
        return super().OnExit()

    def _on_system_loop_timer(self, event):
        # from datetime import datetime
        # print(f'Cengal Timer >> {datetime.now()}>>')
        self._ly()
        if self._idle_for is None:
            self._system_loop_timer.StartOnce(1) # run every 1 millisecond
        else:
            # print(f'Idle for {self._idle_for} milliseconds.')
            idle_for = self._idle_for
            self._idle_for = None
            self._system_loop_timer.StartOnce(idle_for)

    def _on_system_loop_idle(self, next_event_after: Optional[RationalNumber]):
        if next_event_after is None:
            self._idle_for = max(1, sec_to_ms(get_usable_min_sleep_interval()))
        else:
            self._idle_for = max(1, sec_to_ms(next_event_after))


def bind_coro_explicit(wx_entity: Union[App, Frame, Dialog], coro: AnyWorker, *args, **kwargs):
    i: Interface = current_interface()
    coro_id: CoroID = i(PutCoro, coro, *args, **kwargs)
    bind_running_coro(wx_entity, coro_id)


def bind_coro_implicit(wx_entity: Union[App, Frame, Dialog], coro: AnyWorker, *args, **kwargs):
    i: Interface = current_interface()
    coro_id: CoroID = i(PutCoro, cs_coro(coro), *args, **kwargs)
    bind_running_coro(wx_entity, coro_id)


bind_coro = bind_coro_implicit


async def abind_coro_explicit(wx_entity: Union[App, Frame, Dialog], coro: AnyWorker, *args, **kwargs):
    i: Interface = current_interface()
    coro_id: CoroID = await i(PutCoro, coro, *args, **kwargs)
    bind_running_coro(wx_entity, coro_id)


async def abind_coro_implicit(wx_entity: Union[App, Frame, Dialog], coro: AnyWorker, *args, **kwargs):
    i: Interface = current_interface()
    coro_id: CoroID = await i(PutCoro, cs_coro(coro), *args, **kwargs)
    bind_running_coro(wx_entity, coro_id)


abind_coro = abind_coro_implicit


async def asyncio_bind_coro_explicit(wx_entity: Union[App, Frame, Dialog], coro: AnyWorker, *args, **kwargs):
    coro_id: CoroID = await await_task_prim(PutCoro, coro, *args, **kwargs)
    bind_running_coro(wx_entity, coro_id)


async def asyncio_bind_coro_implicit(wx_entity: Union[App, Frame, Dialog], coro: AnyWorker, *args, **kwargs):
    coro_id: CoroID = await await_task_prim(PutCoro, cs_coro(coro), *args, **kwargs)
    bind_running_coro(wx_entity, coro_id)


asyncio_bind_coro = asyncio_bind_coro_implicit


def bind_asyncio_coro(wx_entity: Union[App, Frame, Dialog], awaitable: Awaitable):
    coro_task: Task = create_task(awaitable)
    bind_running_asyncio_coro(wx_entity, coro_task)


def bind_running_coro(wx_entity: Union[App, Frame, Dialog], coro_id: CoroID):
    if isinstance(wx_entity, Window):
        wx_entity = GetTopLevelParent(wx_entity)
    
    if isinstance(wx_entity, App):
        wx_entity._binded_coros.add(coro_id)
    elif isinstance(wx_entity, (Frame, Dialog)):
        def close_binded_coro(event):
            i: Interface = current_interface()
            graceful_coro_destroyer(i, 0.1, coro_id, BindedEntityIsAboutToBeDestroyed)
            event.Skip()
        
        wx_entity.Bind(EVT_CLOSE, close_binded_coro)
    else:
        raise RuntimeError(f'wx_entity must be App, Frame, Dialog or a widget (Window) on a top level window, not {type(wx_entity)}')


def bind_running_asyncio_coro(wx_entity: Union[App, Frame, Dialog], coro_future: Future):
    if isinstance(wx_entity, App):
        wx_entity._binded_coros.add(coro_future)
    elif isinstance(wx_entity, (Frame, Dialog)):
        def close_binded_asyncio_coro(event):
            coro_future.cancel()
            event.Skip()
        
        wx_entity.Bind(EVT_CLOSE, close_binded_asyncio_coro)


def wx_exec_in_coro(func: Callable) -> Callable:
    @wraps(func)
    def wrapper(*args, **kwargs):
        cs: CoroSchedulerType = current_coro_scheduler()
        cs.high_cpu_utilisation_mode = False
        cs.use_internal_sleep = False
        i: Interface = current_interface()
        i(AsyncioLoopRequest().use_higher_level_sleep_manager())
        i(AsyncioLoopRequest().ensure_loop(interrupt_when_no_requests=True))
        i(AsyncioLoopRequest().turn_on_loops_intercommunication())

        app_or_tuple = func(*args, **kwargs)
        on_exit = lambda ret: ret
        if isinstance(app_or_tuple, App):
            app = app_or_tuple
        elif isinstance(app_or_tuple, tuple):
            app, on_exit = app_or_tuple
        else:
            raise RuntimeError("wx_exec_in_coro must return either wx.App or (wx.App, on_exit) tuple")
        
        # Run the event loop
        ret = app.MainLoop()
        ret = i(RunCoro, cs_coro(on_exit), ret)
        i(ShutdownLoop)
        return ret
    
    return cs_coro(wrapper)


def bind_to(wx_entity: Union[App, Frame, Dialog], wrapped_coro: Callable):
    if not (hasattr(wrapped_coro, '__wrapped__') and hasattr(wrapped_coro, '_decorator_func')):
        return wrapped_coro
    
    coro: Union[AnyWorker, Coroutine] = wrapped_coro.__wrapped__
    decorator_func = wrapped_coro._decorator_func
    new_wrapped_coro = decorator_func(coro)
    # if hasattr(wrapped_coro, '__self__'):
    #     new_wrapped_coro = new_wrapped_coro.__get__(wrapped_coro.__self__, wrapped_coro.__self__.__class__)
    
    new_wrapped_coro._binded_entity = wx_entity
    if hasattr(wrapped_coro, '__self__'):
        new_wrapped_coro.__self__ = wrapped_coro.__self__
    
    return new_wrapped_coro


def bind(wrapped_coro: Callable):
    if hasattr(wrapped_coro, '__self__'):
        return bind_to(wrapped_coro.__self__, wrapped_coro)
    else:
        return wrapped_coro 


def event_handler_implicit(coro: AnyWorker):
    def wrapper(*args, **kwargs):
        coro_sign: Signature = signature(coro)
        binded_entity = None
        if 2 == len(coro_sign.parameters):
            # The first parameter is self since event handler must have single parameter: event
            # binded_entity = wrapper.__self__
            binded_entity = args[0]

        i: Interface = current_interface()
        if hasattr(wrapper, '_binded_entity') and hasattr(wrapper, '__self__'):
            args = (wrapper.__self__,) + args
        
        coro_id: CoroID = i(PutCoro, cs_coro(coro), *args, **kwargs)
        if hasattr(wrapper, '_binded_entity'):
            binded_entity = getattr(wrapper, '_binded_entity')
        else:
            if binded_entity is None:
                binded_entity = App.Get()
        
        bind_running_coro(binded_entity, coro_id)
    
    wrapper.__wrapped__ = coro
    wrapper._decorator_func = event_handler_implicit
    return wrapper


event_handler = event_handler_implicit
aevent_handler = event_handler_implicit


def event_handler_explicit(coro: AnyWorker):
    @wraps(coro)
    def wrapper(*args, **kwargs):
        coro_sign: Signature = signature(coro)
        binded_entity = None
        if 3 == len(coro_sign.parameters):
            # The first parameter is self since event handler must have single parameter: event
            # binded_entity = wrapper.__self__
            binded_entity = args[1]

        i: Interface = current_interface()
        if hasattr(wrapper, '_binded_entity') and hasattr(wrapper, '__self__'):
            args = (wrapper.__self__,) + args
        
        coro_id: CoroID = i(PutCoro, coro, *args, **kwargs)
        if hasattr(wrapper, '_binded_entity'):
            binded_entity = getattr(wrapper, '_binded_entity')
        else:
            if binded_entity is None:
                binded_entity = App.Get()
        
        bind_running_coro(binded_entity, coro_id)
    
    wrapper.__wrapped__ = coro
    wrapper._decorator_func = event_handler_explicit
    return wrapper


def asyncio_event_handler(coro: Coroutine):
    @wraps(coro)
    def wrapper(*args, **kwargs):
        coro_sign: Signature = signature(coro)
        binded_entity = None
        if 2 == len(coro_sign.parameters):
            # The first parameter is self since event handler must have single parameter: event
            # binded_entity = wrapper.__self__
            binded_entity = args[0]

        if hasattr(wrapper, '_binded_entity') and hasattr(wrapper, '__self__'):
            args = (wrapper.__self__,) + args
        
        coro_task: Task = create_task(coro(*args, **kwargs))
        if hasattr(wrapper, '_binded_entity'):
            binded_entity = getattr(wrapper, '_binded_entity')
        else:
            if binded_entity is None:
                binded_entity = App.Get()
        
        bind_running_asyncio_coro(binded_entity, coro_task)
    
    wrapper.__wrapped__ = coro
    wrapper._decorator_func = asyncio_event_handler
    return wrapper


def blocking_event_handler_implicit(coro: AnyWorker):
    @wraps(coro)
    def wrapper(*args, **kwargs):
        i: Interface = current_interface()
        if hasattr(wrapper, '_binded_entity') and hasattr(wrapper, '__self__'):
            args = (wrapper.__self__,) + args
        
        i(RunCoro, cs_coro(coro), *args, **kwargs)
    
    wrapper.__wrapped__ = coro
    wrapper._decorator_func = blocking_event_handler_implicit
    return wrapper


def blocking_event_handler_explicit(coro: AnyWorker):
    @wraps(coro)
    def wrapper(*args, **kwargs):
        i: Interface = current_interface()
        if hasattr(wrapper, '_binded_entity') and hasattr(wrapper, '__self__'):
            args = (wrapper.__self__,) + args
        
        i(RunCoro, coro, *args, **kwargs)
    
    wrapper.__wrapped__ = coro
    wrapper._decorator_func = blocking_event_handler_explicit
    return wrapper



#!/usr/bin/env python
# coding=utf-8

# Copyright © 2012-2024 ButenkoMS. All rights reserved. Contacts: <gtalk@butenkoms.space>
# 
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
# 
#     http://www.apache.org/licenses/LICENSE-2.0
# 
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

__all__ = [
        'ValueExistenceNamedTuple',
        'ValueExistence',
        'ValueHolder',
        'ResultExistence',
        'ResultHolder',
        'ErrorExistence',
        'ErrorHolder',
        'ValueCache',
        'ValueTypeNamedTuple',
        'ValueType',
        'ValueWithTypeNamedTuple',
        'ValueWithType',
    ]


from cengal.introspection.inspect import entity_module_importable_str_and_owning_names_path, entity_by_name_module_importable_str_and_owning_names_path
from collections import namedtuple
from collections.abc import Sequence
from typing import Any, TypeVar, Generic, Tuple, NamedTuple, Union, Optional

"""
Module Docstring
Docstrings: http://www.python.org/dev/peps/pep-0257/
"""

__author__ = "ButenkoMS <gtalk@butenkoms.space>"
__copyright__ = "Copyright © 2012-2024 ButenkoMS. All rights reserved. Contacts: <gtalk@butenkoms.space>"
__credits__ = ["ButenkoMS <gtalk@butenkoms.space>", ]
__license__ = "Apache License, Version 2.0"
__version__ = "4.4.0"
__maintainer__ = "ButenkoMS <gtalk@butenkoms.space>"
__email__ = "gtalk@butenkoms.space"
# __status__ = "Prototype"
__status__ = "Development"
# __status__ = "Production"


# TODO: implement Cython version (smart_values are used by our Cython code)


VT = TypeVar('VT')


ValueExistenceNamedTuple = namedtuple('ValueExistenceNamedTuple', ['existence', 'value'])


class ValueExistence(Generic[VT], Sequence):
    __slots__ = ('existence', '_value')

    def __init__(self, existence: bool = False, value: VT = None):
        self.existence: bool = existence
        self._value: VT = value

    def __len__(self):
        return 2

    def __getitem__(self, index):
        if 0 == index:
            return self.existence
        elif 1 == index:
            return self._value
        else:
            raise IndexError
    
    @property
    def value(self):
        return self._value
    
    @value.setter
    def value(self, value):
        self.existence = True
        self._value = value

    def __bool__(self):
        return self.existence

    def __nonzero__(self):
        return self.__bool__()

    def __str__(self):
        return f'{type(self).__name__}({repr(self.existence)}, {repr(self.value)})'

    def __repr__(self):
        return self.__str__()

    def __getstate__(self):
        return self.existence, self.value

    def __setstate__(self, state):
        existence, value = state
        self.value = value
        self.existence = existence
    
    def __eq__(self, __value: object) -> bool:
        if isinstance(__value, ValueExistence):
            return (self.existence == __value.existence) and (self.value == __value.value)
        else:
            if self.existence:
                return self.value == __value
            else:
                return False
    
    def __ne__(self, __value: object) -> bool:
        return not self.__eq__(__value)
    
    def to_namedtuple(self) -> ValueExistenceNamedTuple:
        return ValueExistenceNamedTuple(self.existence, self.value)
    
    def to_tuple(self) -> Tuple:
        return self.existence, self.value
    
    def to_dict(self) -> dict:
        return {
            'existence': self.existence,
            'value': self.value,
        }
    
    def serialize_to_dict(self) -> Tuple[dict, dict]:
        module_importable_str, owning_names_path = entity_module_importable_str_and_owning_names_path(self.__class__)
        return {
            'existence': self.existence,
            'value': self.value,
        }, {
            'class_name': self.__class__,
            'module_importable_str': module_importable_str,
            'owning_names_path': owning_names_path,
        }
    
    @classmethod
    def from_namedtuple(cls, named_tuple: ValueExistenceNamedTuple) -> 'ValueExistence':
        return cls(named_tuple.existence, named_tuple.value)
    
    @classmethod
    def from_tuple(cls, tuple_: Tuple) -> 'ValueExistence':
        return cls(*tuple_)
    
    @classmethod
    def from_dict(cls, dict_: dict) -> 'ValueExistence':
        return cls(dict_['existence'], dict_['value'])
    
    @classmethod
    def deserialize_from_dict(cls, dict_: dict, owning_info: Optional[str] = None) -> 'ValueExistence':
        if owning_info is None:
            return cls(dict_['existence'], dict_['value'])
        else:
            class_type = entity_by_name_module_importable_str_and_owning_names_path(owning_info['class_name'], owning_info['module_importable_str'], owning_info['owning_names_path'])
            return class_type(dict_['existence'], dict_['value'])
    
    @classmethod
    def from_other(cls, value: Union['ValueExistence', ValueExistenceNamedTuple, Tuple, VT]) -> 'ValueExistence':
        if isinstance(value, ValueExistence):
            return cls(value.existence, value.value)
        elif isinstance(value, ValueExistenceNamedTuple):
            return cls.from_namedtuple(value)
        elif isinstance(value, tuple):
            return cls.from_tuple(value)
        elif isinstance(value, dict):
            return cls.from_dict(value)
        else:
            try:
                return cls(value.existence, value.value)
            except AttributeError:
                return cls(True, value)


class ValueHolder(ValueExistence[VT]):
    def __init__(self, existence: bool = False, value: VT = None):
        super().__init__(existence, value)


class ResultExistence(ValueExistence[VT]):
    def __init__(self, existence: bool = False, value: VT = None):
        super().__init__(existence, value)


class ResultHolder(ValueExistence[VT]):
    def __init__(self, existence: bool = False, value: VT = None):
        super().__init__(existence, value)


class ErrorExistence(ValueExistence[VT]):
    def __init__(self, existence: bool = False, value: VT = None):
        super().__init__(existence, value)


class ErrorHolder(ValueExistence[VT]):
    def __init__(self, existence: bool = False, value: VT = None):
        super().__init__(existence, value)


class ValueCache(ValueExistence[VT]):
    __slots__ = tuple()

    def __init__(self, existence: bool = False, value: VT = None):
        super(ValueCache, self).__init__(existence, value)

    def __call__(self, *args, **kwargs):
        self.existence = False

    def get(self) -> VT:
        return self.value

    def set(self, new_value: VT):
        self.existence = True
        self.value = new_value

    def reset(self):
        self.existence = False

    def __getstate__(self):
        return self.existence, self.value

    def __setstate__(self, state):
        existence, value = state
        self.value = value
        self.existence = existence
    
    def __eq__(self, __value: object) -> bool:
        if isinstance(__value, ValueExistence):
            return (self.existence == __value.existence) and (self.value == __value.value)
        else:
            if self.existence:
                return self.value == __value
            else:
                return False
    
    def __ne__(self, __value: object) -> bool:
        return not self.__eq__(__value)


TT = TypeVar('TT')


ValueTypeNamedTuple = namedtuple('ValueTypeNamedTuple', ['type_id', 'value'])


class ValueType(Generic[TT, VT], Sequence):
    __slots__ = ('type_id', 'value')

    def __init__(self, type_id: TT, value: VT):
        self.type_id: TT = type_id
        self.value: VT = value

    def __len__(self):
        return 2

    def __getitem__(self, index):
        if 0 == index:
            return self.type_id
        elif 1 == index:
            return self.value
        else:
            raise IndexError

    def __eq__(self, other):
        # "__ne__() delegates to __eq__() and inverts the value"
        if isinstance(other, (ValueType, ValueWithType)):
            return self.type_id == other.type_id
        else:
            return self.type_id == other

    def __getstate__(self):
        return self.type_id, self.value

    def __setstate__(self, state):
        self.type_id, self.value = state

    def __str__(self):
        return f'{type(self).__name__}({repr(self.type_id)}, {repr(self.value)})'

    def __repr__(self):
        return self.__str__()
    
    def to_namedtuple(self) -> ValueTypeNamedTuple:
        return ValueTypeNamedTuple(self.type_id, self.value)
    
    def to_tuple(self) -> Tuple:
        return self.type_id, self.value
    
    def to_dict(self) -> dict:
        return {
            'type_id': self.type_id,
            'value': self.value,
        }
    
    def serialize_to_dict(self) -> Tuple[dict, dict]:
        module_importable_str, owning_names_path = entity_module_importable_str_and_owning_names_path(self.__class__)
        return {
            'type_id': self.type_id,
            'value': self.value,
        }, {
            'class_name': self.__class__,
            'module_importable_str': module_importable_str,
            'owning_names_path': owning_names_path,
        }
    
    @classmethod
    def from_namedtuple(cls, named_tuple: ValueTypeNamedTuple) -> 'ValueExistence':
        return cls(named_tuple.type_id, named_tuple.value)
    
    @classmethod
    def from_tuple(cls, tuple_: Tuple) -> 'ValueExistence':
        return cls(*tuple_)
    
    @classmethod
    def from_dict(cls, dict_: dict) -> 'ValueExistence':
        return cls(dict_['type_id'], dict_['value'])
    
    @classmethod
    def deserialize_from_dict(cls, dict_: dict, owning_info: Optional[str] = None) -> 'ValueExistence':
        if owning_info is None:
            return cls(dict_['type_id'], dict_['value'])
        else:
            class_type = entity_by_name_module_importable_str_and_owning_names_path(owning_info['class_name'], owning_info['module_importable_str'], owning_info['owning_names_path'])
            return class_type(dict_['type_id'], dict_['value'])
    
    @classmethod
    def from_other(cls, value: Union['ValueExistence', ValueTypeNamedTuple, Tuple, VT]) -> 'ValueExistence':
        if isinstance(value, ValueExistence):
            return cls(value.type_id, value.value)
        elif isinstance(value, ValueTypeNamedTuple):
            return cls.from_namedtuple(value)
        elif isinstance(value, tuple):
            return cls.from_tuple(value)
        elif isinstance(value, dict):
            return cls.from_dict(value)
        else:
            try:
                return cls(value.type_id, value.value)
            except AttributeError:
                return cls(None, value)


ValueWithTypeNamedTuple = namedtuple('ValueWithTypeNamedTuple', ['type_id', 'value'])


class ValueWithType(Generic[TT, VT], Sequence):
    __slots__ = ('type_id', 'value')

    def __init__(self, type_id: TT, value: VT):
        self.type_id: TT = type_id
        self.value: VT = value

    def __len__(self):
        return 2

    def __getitem__(self, index):
        if 0 == index:
            return self.type_id
        elif 1 == index:
            return self.value
        else:
            raise IndexError

    def __eq__(self, other):
        # "__ne__() delegates to __eq__() and inverts the value"
        if isinstance(other, ValueType):
            return self.type_id == other.type_id
        elif isinstance(other, ValueWithType):
            return (self.type_id == other.type_id) and (self.value == other.value)
        else:
            return self.value == other

    def __getstate__(self):
        return self.type_id, self.value

    def __setstate__(self, state):
        self.type_id, self.value = state

    def __str__(self):
        return f'{type(self).__name__}({repr(self.type_id)}, {repr(self.value)})'

    def __repr__(self):
        return self.__str__()
    
    def to_namedtuple(self) -> ValueWithTypeNamedTuple:
        return ValueWithTypeNamedTuple(self.type_id, self.value)
    
    def to_tuple(self) -> Tuple:
        return self.type_id, self.value
    
    def to_dict(self) -> dict:
        return {
            'type_id': self.type_id,
            'value': self.value,
        }
    
    def serialize_to_dict(self) -> Tuple[dict, dict]:
        module_importable_str, owning_names_path = entity_module_importable_str_and_owning_names_path(self.__class__)
        return {
            'type_id': self.type_id,
            'value': self.value,
        }, {
            'class_name': self.__class__,
            'module_importable_str': module_importable_str,
            'owning_names_path': owning_names_path,
        }
    
    @classmethod
    def from_namedtuple(cls, named_tuple: ValueWithTypeNamedTuple) -> 'ValueExistence':
        return cls(named_tuple.type_id, named_tuple.value)
    
    @classmethod
    def from_tuple(cls, tuple_: Tuple) -> 'ValueExistence':
        return cls(*tuple_)
    
    @classmethod
    def from_dict(cls, dict_: dict) -> 'ValueExistence':
        return cls(dict_['type_id'], dict_['value'])
    
    @classmethod
    def deserialize_from_dict(cls, dict_: dict, owning_info: Optional[str] = None) -> 'ValueExistence':
        if owning_info is None:
            return cls(dict_['type_id'], dict_['value'])
        else:
            class_type = entity_by_name_module_importable_str_and_owning_names_path(owning_info['class_name'], owning_info['module_importable_str'], owning_info['owning_names_path'])
            return class_type(dict_['type_id'], dict_['value'])
    
    @classmethod
    def from_other(cls, value: Union['ValueExistence', ValueWithTypeNamedTuple, Tuple, VT]) -> 'ValueExistence':
        if isinstance(value, ValueExistence):
            return cls(value.type_id, value.value)
        elif isinstance(value, ValueWithTypeNamedTuple):
            return cls.from_namedtuple(value)
        elif isinstance(value, tuple):
            return cls.from_tuple(value)
        elif isinstance(value, dict):
            return cls.from_dict(value)
        else:
            try:
                return cls(value.type_id, value.value)
            except AttributeError:
                return cls(None, value)



#!/usr/bin/env python
# coding=utf-8

# Copyright © 2012-2024 ButenkoMS. All rights reserved. Contacts: <gtalk@butenkoms.space>
# 
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
# 
#     http://www.apache.org/licenses/LICENSE-2.0
# 
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from contextlib import contextmanager
from typing import List, Tuple, Optional, Any, Dict, cast

"""
Module Docstring
Docstrings: http://www.python.org/dev/peps/pep-0257/
"""

__author__ = "ButenkoMS <gtalk@butenkoms.space>"
__copyright__ = "Copyright © 2012-2024 ButenkoMS. All rights reserved. Contacts: <gtalk@butenkoms.space>"
__credits__ = ["ButenkoMS <gtalk@butenkoms.space>", ]
__license__ = "Apache License, Version 2.0"
__version__ = "4.4.0"
__maintainer__ = "ButenkoMS <gtalk@butenkoms.space>"
__email__ = "gtalk@butenkoms.space"
# __status__ = "Prototype"
__status__ = "Development"
# __status__ = "Production"

EntityExactHolder = Any
EntityHolder = Optional[Any]
EntityName = str
OriginalExactEntity = Tuple[EntityExactHolder, EntityName]
OriginalEntity = Tuple[EntityHolder, EntityName]
EntityMock = Any

@contextmanager
def patch_entity(original: OriginalExactEntity, mock: EntityMock):
    holder, name = original
    buff_original_value = getattr(holder, name)
    setattr(holder, name, mock)
    try:
        yield buff_original_value
    finally:
        setattr(holder, name, buff_original_value)

@contextmanager
def patch_builtins(name: EntityName, mock: EntityMock):
    buff_original_value = __builtins__[name]
    __builtins__[name] = mock
    try:
        yield buff_original_value
    finally:
        __builtins__[name] = buff_original_value

@contextmanager
def patch(original: OriginalEntity, mock: EntityMock):
    holder, name = original
    if holder is None:
        return patch_builtins(name, mock)
    else:
        return patch_entity(cast(OriginalExactEntity, original), mock)
        
@contextmanager
def patch_set(patch_set: Dict[OriginalEntity, EntityMock]):
    buff = dict()
    for original, mock in patch_set.items():
        original: OriginalEntity = original
        mock: EntityMock = mock
        holder, name = original
        holder: EntityHolder = holder
        name: EntityName = name
        if holder is None:
            buff[original] = __builtins__[name]
            __builtins__[name] = mock
        else:
            buff[original] = getattr(holder, name)
            setattr(holder, name, mock)
    try:
        yield buff
    finally:
        for original, mock in patch_set.items():
            holder, name = original
            if holder is None:
                __builtins__[name] = buff[original]
            else:
                setattr(holder, name, buff[original])
        


#!/usr/bin/env python
# coding=utf-8

# Copyright © 2012-2024 ButenkoMS. All rights reserved. Contacts: <gtalk@butenkoms.space>
# 
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
# 
#     http://www.apache.org/licenses/LICENSE-2.0
# 
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


"""
Module Docstring
Docstrings: http://www.python.org/dev/peps/pep-0257/
"""


__author__ = "ButenkoMS <gtalk@butenkoms.space>"
__copyright__ = "Copyright © 2012-2024 ButenkoMS. All rights reserved. Contacts: <gtalk@butenkoms.space>"
__credits__ = ["ButenkoMS <gtalk@butenkoms.space>", ]
__license__ = "Apache License, Version 2.0"
__version__ = "4.4.0"
__maintainer__ = "ButenkoMS <gtalk@butenkoms.space>"
__email__ = "gtalk@butenkoms.space"
# __status__ = "Prototype"
__status__ = "Development"
# __status__ = "Production"


# from distutils.dist import Distribution
from os import environ

from setuptools._distutils.dist import Distribution

from cengal.file_system.path_manager import path_relative_to_src, RelativePath, get_relative_path_part, sep
from cengal.file_system.directory_manager import current_src_dir
from cengal.file_system.directory_manager import filtered_file_list, FilteringType, filtered_file_list_traversal, file_list_traversal, FilteringEntity, \
    change_current_dir, secure_current_dir
from cengal.build_tools.prepare_cflags import prepare_cflags, concat_cflags, prepare_compile_time_env, prepare_cflags_dict, prepare_given_cflags_dict, \
    prepare_pyx_flags_dict
from cengal.introspection.inspect import get_exception, entity_repr_limited_try_qualname, pifrl, pdi, class_properties_values_including_overrided
from cengal.hardware.info.cpu import cpu_info
from cengal.system import OS_TYPE, TEMPLATE_MODULE_NAME
from shutil import rmtree
from os import remove, getcwd
from os.path import splitext, normpath, join as path_join, basename, split
from setuptools import Extension as SetuptoolsExtension
from Cython.Distutils import Extension as CythonExtension
from distutils.command.build import build as build_orig
from distutils.command.build_ext import build_ext as build_ext_orig
from setuptools.command.sdist import sdist as sdist_orig
import json
import importlib

from os.path import isdir, exists, isfile

import setuptools
import platform

from cengal.file_system.path_manager import RelativePath, get_relative_path_part
from cengal.file_system.directory_manager import current_src_dir
from cengal.file_system.directory_manager import file_list_traversal, FilteringEntity
from cengal.os.execute import prepare_params
from setuptools.discovery import find_package_path
import subprocess
from cengal.build_tools.build_extensions import *
from pprint import pprint
from typing import List, Dict, Optional, Iterable, Callable, Sequence, Tuple, Union, Type, Set


_PYTHON_VERSION = platform.python_version_tuple()
BUILD_CONFIG_FILENAME: str = '__build_config.py'
GITKEEP_FILE_NAME = '.gitkeep'
MANIFEST_INCLUDED_FILES_FILE_NAME = 'MANIFEST.in'
MANIFEST_INCLUDED_FILES_TEMPLATE_FILE_NAME = 'MANIFEST.in.template'


cython_file_ext = '.pyx'
cython_transpiled_ext = {'.c'}
compilable_ext = {'.pyx', '.cpp', '.c++', '.cxx', '.cc'}
codegen_files_ext = {'.c', '.cpp', '.c++', '.cxx', '.cc'}
headers_ext = {'.h', '.hpp', '.h++', '.hh', '.hxx'}
libs_ext = {'.lib', '.dylib', '.dll', '.so', '.exp', '.obj'}
all_ext = cython_transpiled_ext | compilable_ext | headers_ext | libs_ext


def find_good_packages(where: str = '.', exclude: Iterable[str] = tuple(), include: Iterable[str] = ('*',), 
                       modules_to_ignore: Iterable[str] = tuple(), python_2_modules: Iterable[str] = tuple()):
    all_packages = setuptools.find_packages(where, exclude, include)
    good_packages = list()
    for package in all_packages:
        is_good = True
        for item in modules_to_ignore:
            if package.startswith(item):
                is_good = False
                break

        if '3' == _PYTHON_VERSION[0]:
            for item in python_2_modules:
                if package.startswith(item):
                    is_good = False
                    break

        if is_good:
            good_packages.append(package)

    return good_packages


def find_package_data_filter(filtering_entity: FilteringEntity, data):
    if FilteringEntity.dirpath == filtering_entity:
        return True
    elif FilteringEntity.dirname == filtering_entity:
        return True
    elif FilteringEntity.filename == filtering_entity:
        dirpath, filename = data
        return GITKEEP_FILE_NAME != filename
    elif FilteringEntity.aggregated == filtering_entity:
        return data
    else:
        raise NotImplementedError()
        

def find_package_data(package_src_relative_path: str, good_packages: List[str], depth: Optional[int] = 1, root_rel: Optional[RelativePath] = None) -> Tuple[Dict[str, List[str]], List[str]]:
    depth = depth or 0
    depth += 1
    root_rel = RelativePath(current_src_dir()) if root_rel is None else root_rel
    root_path = root_rel('')
    print(f'{root_path=}')
    package_src_rel = RelativePath(root_rel(package_src_relative_path))
    package_src_path = package_src_rel('')
    print(f'{package_src_path=}')

    packages_data_dict: Dict[str, List[str]] = dict()
    manifest_included_files: Set[str] = set()
    for package_name in good_packages:
        package_path = find_package_path(package_name, dict(), '')
        package_full_path: str = root_rel(package_path)
        package_full_path_rel: RelativePath = RelativePath(package_full_path)
        package_data: Set[str] = set()
        possible_data_dir: str = package_full_path_rel('data')
        if exists(possible_data_dir) and isdir(possible_data_dir):
            possible_data_dir_rel = RelativePath(possible_data_dir)
            travers_result = file_list_traversal(possible_data_dir, find_package_data_filter, remove_empty_dirpaths=False)
            for dirpath, dirnames, filenames in travers_result:
                dirpath_rel = RelativePath(dirpath)
                for filename in filenames:
                    file_full_path = dirpath_rel(filename)
                    package_data.add(get_relative_path_part(file_full_path, package_full_path))
                    manifest_included_files.add(get_relative_path_part(file_full_path, root_path))
        
        pyx_files = filtered_file_list_traversal(package_full_path, FilteringType.including, {'.pyx'}, remove_empty_items=True)
        for dirpath, dirnames, filenames in pyx_files:
            dirpath_rel = RelativePath(dirpath)
            for filename in filenames:
                file_full_path = dirpath_rel(filename)
                package_data.add(get_relative_path_part(file_full_path, package_full_path))
                manifest_included_files.add(get_relative_path_part(file_full_path, root_path))
        
        if package_data:
            packages_data_dict[package_name] = list(package_data)

    return packages_data_dict, list(manifest_included_files)


def generate_manifest_included_files(manifest_included_files: List[str], root_rel: RelativePath):
    manifest_included_files_template_path: str = root_rel(MANIFEST_INCLUDED_FILES_TEMPLATE_FILE_NAME)
    manifest_included_files_template_contents: str = str()
    if exists(manifest_included_files_template_path) and isfile(manifest_included_files_template_path):
        with open(manifest_included_files_template_path, 'r') as f:
            manifest_included_files_template_contents = f.read()

    manifest_included_files_path: str = root_rel(MANIFEST_INCLUDED_FILES_FILE_NAME)
    with open(manifest_included_files_path, 'w') as f:
        if manifest_included_files_template_contents:
            f.write(manifest_included_files_template_contents + '\n')
        
        for file in manifest_included_files:
            f.write(f'include {file}\n')


def get_file_name_without_extension(path):
    file_name = basename(path)  # Get the base name from the path
    file_name_without_extension = splitext(file_name)[0]  # Split the file name and extension, and take only the file name part
    return file_name_without_extension


def remove_header_files(ext_modules: Sequence[CythonExtension]) -> Sequence[CythonExtension]:
    for ext_module in ext_modules:
        if isinstance(ext_module, CythonExtension):
            new_sources = list()
            for source in ext_module.sources:
                _, file_extension = splitext(source)
                if file_extension not in headers_ext:
                    new_sources.append(source)
            
            ext_module.sources = new_sources
    
    return ext_modules


def extend_file_names_to_root_relative_paths(root_path: str, dir_path_obj: RelativePath, file_names_or_path: Sequence[str]) -> Sequence[str]:
    result = list()
    for file_name_or_path in file_names_or_path:
        if not file_name_or_path:
            continue
        
        owner_path, file_name = split(file_name_or_path)
        file_name = file_name.strip()
        if owner_path:
            result.append(file_name_or_path)
        else:
            result.append(path_join('.', get_relative_path_part(dir_path_obj(file_name), root_path)))
    
    return result


def process_macros(ext_modules: Sequence[CythonExtension]) -> Sequence[CythonExtension]:
    for ext_module in ext_modules:
        if isinstance(ext_module, CythonExtension):
            if hasattr(ext_module, 'compile_time_env'):
                compile_time_env = ext_module.compile_time_env
            else:
                compile_time_env = dict()
            
            if hasattr(ext_module, 'cython_compile_time_env'):
                cython_compile_time_env = ext_module.cython_compile_time_env or dict()
                compile_time_env.update(cython_compile_time_env)
                ext_module.cython_compile_time_env = dict()
            
            ext_module.compile_time_env = compile_time_env

            if hasattr(ext_module, 'define_macros'):
                define_macros = ext_module.define_macros
            else:
                define_macros = list()
            
            if isinstance(define_macros, dict):
                define_macros = list(define_macros.items())
            
            new_define_macros = list()
            for macros_name, macros_value in define_macros:
                macros_value = str(macros_value)
                macros_value = macros_value.strip()
                if not macros_value:
                    macros_value = None
                
                new_define_macros.append((macros_name, macros_value))
            
            ext_module.define_macros = define_macros

    return ext_modules


def find_and_prepare_cython_modules(package_src_relative_path: str, additional_cflags: Dict[str, Tuple[bool, Union[str, int]]], depth = 1, root_rel: Optional[RelativePath] = None):
    with secure_current_dir():
        depth = depth or 0
        depth += 1
        root_rel = RelativePath(current_src_dir(depth)) if root_rel is None else root_rel
        root_path = root_rel('')
        package_src_rel = RelativePath(root_rel(package_src_relative_path))
        package_src_path = package_src_rel('')

        travers_result = filtered_file_list_traversal(package_src_path, FilteringType.off, None, remove_empty_items=True, use_spinner=False)
        for dir_path, dirs, files in travers_result:
            if '__pycache__' in dirs:
                pycache_path = RelativePath(dir_path)('__pycache__')
                try:
                    rmtree(pycache_path)
                except OSError as e:
                    print(get_exception())


        def filter(filtering_entity: FilteringEntity, data):
            if FilteringEntity.dirpath == filtering_entity:
                # Ignore package_src/_template_module
                dirpath, dirnames, filenames = data
                rel_path = get_relative_path_part(dirpath, package_src_path)
                rel_path = rel_path.strip()
                if rel_path:
                    rel_path_parts = rel_path.split(sep)
                    if rel_path_parts and TEMPLATE_MODULE_NAME == rel_path_parts[0]:
                        return False
                
                return True
            elif FilteringEntity.dirname == filtering_entity:
                return True
            elif FilteringEntity.filename == filtering_entity:
                dirpath, filename = data
                if BUILD_CONFIG_FILENAME == filename:
                    return True
                
                file_name, file_extension = splitext(filename)
                if file_extension in all_ext:
                    return True
                
                return False
            elif FilteringEntity.aggregated == filtering_entity:
                return data
            else:
                raise NotImplementedError()
            
        travers_result = file_list_traversal(package_src_path, filter, remove_empty_dirpaths=True)
        # travers_result = filtered_file_list_traversal(package_src_path, FilteringType.including, {'.pyx', '.c', '.lib', '.dll', '.so'}, remove_empty_items=True, use_spinner=False)
        
        result = list()
        for dir_path, dirs, files in travers_result:
            # Ignore package_src/_template_module
            rel_path = get_relative_path_part(dir_path, package_src_path)
            rel_path = rel_path.strip()
            if rel_path:
                rel_path_parts = rel_path.split(sep)
                if rel_path_parts and TEMPLATE_MODULE_NAME == rel_path_parts[0]:
                    continue

            dir_path_obj = RelativePath(dir_path)
            extensions = dict()
            for file in files:
                filename, file_extension = splitext(file)
                if file_extension not in extensions:
                    extensions[file_extension] = set()
                
                extensions[file_extension].add(file)
            
            is_exctension: bool = False
            build_config: dict = None
            if BUILD_CONFIG_FILENAME in files:
                build_config_module_name, _ = splitext(BUILD_CONFIG_FILENAME)
                build_config_full_path: str = dir_path_obj(build_config_module_name)
                name = get_relative_path_part(build_config_full_path, root_rel(''))
                name_parts = name.split(sep)
                module_full_name = '.'.join([name_part for name_part in name_parts if name_part])
                build_config_module = importlib.import_module(module_full_name)
                build_config = build_config_module.build_config()
                is_exctension = True

            sub_result = list()
            if (cython_file_ext in extensions) or (build_config is not None):
                for file_extension, files in extensions.items():
                    for file in files:
                        if file_extension.lower() in (compilable_ext | headers_ext):
                            is_exctension = True
                            sub_result.append(path_join('.', get_relative_path_part(dir_path_obj(file), root_path)))
                        
                        if file_extension.lower() in codegen_files_ext:
                            filename, _ = splitext(file)
                            if filename.endswith('__cython') or filename.endswith('__compiled') or filename.endswith('__python'):
                                try:
                                    remove(dir_path_obj(file))
                                except OSError as e:
                                    print(get_exception())
                            else:
                                is_exctension = True
                                sub_result.append(path_join('.', get_relative_path_part(dir_path_obj(file), root_path)))
                        
                        if file_extension.lower() in libs_ext:
                            try:
                                remove(dir_path_obj(file))
                            except OSError as e:
                                print(get_exception())
            
            if is_exctension:
                name = get_relative_path_part(dir_path_obj(''), root_rel(''))
                name_parts = name.split(sep)
                if build_config is None:
                    if 1 == len(sub_result) and sub_result[0].endswith(cython_file_ext):
                        name_parts.append(get_file_name_without_extension(sub_result[0]))
                    else:
                        name_parts.append('cython_module')
                    
                    name = '.'.join([name_part for name_part in name_parts if name_part])
                    extension: CythonExtension = CythonExtension(
                        name, 
                        sources=list(set(sub_result)),
                        language="c",
                        # cython_compile_time_env=prepare_cflags_dict(additional_cflags),
                        define_macros=prepare_cflags_dict(additional_cflags),
                    )
                else:
                    if isinstance(build_config, (dict, tuple)):
                        extension_type = CythonExtension
                        if isinstance(build_config, tuple):
                            extension_type, build_config = build_config
                            extension_type = extension_type if extension_type is not None else CythonExtension
                        
                        if 'Windows' == OS_TYPE:
                            # CythonExtension :raises setuptools.errors.PlatformError: if 'runtime_library_dirs' is specified on Windows. (since v63)
                            build_config.pop('runtime_library_dirs', None)
                        
                        name_parts.append(build_config['name'])
                        name = '.'.join([name_part for name_part in name_parts if name_part])
                        build_config.pop('name', None)

                        cython_compile_time_env=prepare_cflags_dict(additional_cflags)
                        if 'cython_compile_time_env' in build_config:
                            cython_compile_time_env.update(build_config['cython_compile_time_env'])
                            build_config.pop('cython_compile_time_env', None)

                        define_macros=prepare_cflags_dict(additional_cflags)
                        if 'define_macros' in build_config:
                            define_macros.update(build_config['define_macros'])
                            build_config.pop('define_macros', None)
                        
                        sources = extend_file_names_to_root_relative_paths(root_path, dir_path_obj, build_config.pop('sources', list()))
                        sources.extend(sub_result)
                        
                        extension: CythonExtension = extension_type(
                            name, 
                            sources=list(set(sources)),
                            # cython_compile_time_env=cython_compile_time_env,
                            define_macros=define_macros,
                            **build_config
                        )
                    elif isinstance(build_config, CengalBuildExtension):
                        build_config.path = dir_path_obj(BUILD_CONFIG_FILENAME)
                        build_config.module_import_str = module_full_name
                        build_config.files.extend(sub_result)
                        extension = build_config
                    else:
                        raise RuntimeError(f'Unknown build_config type: {type(build_config)}')
                
                result.append(extension)
            else:
                result.extend(sub_result)
        
        return result


class BuildConfig:
    def __init__(self) -> None:
        self.package_build_is_in_debug_mode: str = str()
        self.package_build_is_in_sdist_mode: str = str()
        self.additional_pyx_flags: Dict[str, Tuple[bool, Union[str, int]]] = dict()
        self.find_package_data: Callable = None
        self.root_rel: RelativePath = None


class sdist(sdist_orig):
    def __init__(self, dist: Distribution, build_config: BuildConfig) -> None:
        self.build_config: BuildConfig = build_config
        super().__init__(dist)

    def run(self):
        # print('<<< sdist command is currently being run >>>')
        with secure_current_dir():
            if self.build_config.package_build_is_in_debug_mode in environ:
                import debugpy
                debugpy.breakpoint()
            
            # print("sdist command is currently being run")
            environ[self.build_config.package_build_is_in_sdist_mode] = 'True'
            packages_data_dict, manifest_included_files = self.build_config.find_package_data()
            # from pprint import pprint
            # print()
            # print('<<< packages_data_dict: >>>')
            # pprint(packages_data_dict)
            # print()
            # print('<<< manifest_included_files: >>>')
            # pprint(manifest_included_files)
            # print()
            generate_manifest_included_files(manifest_included_files, self.build_config.root_rel)
        
        super().run()


class build(build_orig):
    def __init__(self, dist: Distribution, build_config: BuildConfig) -> None:
        self.build_config: BuildConfig = build_config
        super().__init__(dist)

    def finalize_options(self):
        # print('<<< build.finalize_options is currently being run >>>')
        try:
            with secure_current_dir():
                setuptools_extensions = [ext for ext in self.distribution.ext_modules if isinstance(ext, SetuptoolsExtension)]
                cython_extensions = [ext for ext in self.distribution.ext_modules if isinstance(ext, CythonExtension)]
                cengal_extensions = (ext for ext in self.distribution.ext_modules if (isinstance(ext, CengalBuildExtension) and (not ext.store_as_data)))
                cengal_result_extensions = [ext() for ext in cengal_extensions]
                setuptools_extensions.extend([ext for ext in cengal_result_extensions if isinstance(ext, SetuptoolsExtension)])
                cython_extensions.extend([ext for ext in cengal_result_extensions if isinstance(ext, CythonExtension)])
                cengal_extensions_store_as_data = (ext for ext in self.distribution.ext_modules if (isinstance(ext, CengalBuildExtension) and ext.store_as_data))

                if self.build_config.package_build_is_in_debug_mode in environ:
                    import debugpy
                    debugpy.breakpoint()
            
            super().finalize_options()

            with secure_current_dir():
                package_data = self.distribution.package_data or dict()
                
                if self.build_config.package_build_is_in_sdist_mode in environ:
                    for ext in cengal_extensions_store_as_data:
                        files = ext.sdist()
                        if files is None:
                            continue

                        if ext.package not in package_data:
                            package_data[ext.package] = list()
                        
                        package_data[ext.package].extend(files)
                    
                    result_ext_modules = list()
                    result_ext_modules.extend(cython_extensions)
                    result_ext_modules.extend(setuptools_extensions)
                    self.distribution.package_data = package_data
                    self.distribution.ext_modules = result_ext_modules
                    # from pprint import pprint
                    # print()
                    # print('<<< self.distribution.package_data: >>>')
                    # pprint(self.distribution.package_data)
                    # print()
                    # print('<<< self.distribution.ext_modules: >>>')
                    # pprint(self.distribution.ext_modules)
                    # print()
                else:
                    for ext in cengal_extensions_store_as_data:
                        files = ext()
                        if files is None:
                            continue

                        if ext.package not in package_data:
                            package_data[ext.package] = list()
                        
                        package_data[ext.package].extend(files)

                    from Cython.Build import cythonize
                    result_ext_modules = list()
                    try:
                        cwd_before_cythonize = getcwd()
                        pyx_flags_dict = prepare_pyx_flags_dict(self.build_config.additional_pyx_flags)
                        result_ext_modules.extend(remove_header_files(
                            cythonize(process_macros(cython_extensions),
                            compiler_directives={'language_level': '3'},
                            compile_time_env = pyx_flags_dict,
                            )))
                    except:
                        print(f'DEBUG: {cwd_before_cythonize=} | {getcwd()=}')
                        for extension in cython_extensions:
                            pprint(class_properties_values_including_overrided(extension))
                        
                        raise
                    
                    result_ext_modules.extend(setuptools_extensions)
                    for ext in result_ext_modules:
                        if not hasattr(ext, 'extra_compile_args'):
                            ext.extra_compile_args = list()
                        
                        if ext.extra_compile_args is None:
                            ext.extra_compile_args = list()
                        
                        if 'Darwin' == OS_TYPE:
                            if cpu_info().is_x86:
                                ext.extra_compile_args += ['-arch', 'x86_64']
                            elif cpu_info().is_arm:
                                ext.extra_compile_args += ['-arch', 'arm64']
                    
                    self.distribution.package_data = package_data
                    self.distribution.ext_modules = result_ext_modules

                if self.build_config.package_build_is_in_debug_mode in environ:
                    debugpy.breakpoint()
                    print()
        finally:
            # print('<<< build.finalize_options is done >>>')
            pass


class build_ext(build_ext_orig):
    def __init__(self, dist: Distribution, build_config: BuildConfig) -> None:
        self.build_config: BuildConfig = build_config
        super().__init__(dist)

    def finalize_options(self):
        with secure_current_dir():
            setuptools_extensions = [ext for ext in self.distribution.ext_modules if isinstance(ext, SetuptoolsExtension)]
            cython_extensions = [ext for ext in self.distribution.ext_modules if isinstance(ext, CythonExtension)]
            cengal_extensions = (ext for ext in self.distribution.ext_modules if (isinstance(ext, CengalBuildExtension) and (not ext.store_as_data)))
            cengal_result_extensions = [ext() for ext in cengal_extensions]
            setuptools_extensions.extend([ext for ext in cengal_result_extensions if isinstance(ext, SetuptoolsExtension)])
            cython_extensions.extend([ext for ext in cengal_result_extensions if isinstance(ext, CythonExtension)])
            cengal_extensions_store_as_data = (ext for ext in self.distribution.ext_modules if (isinstance(ext, CengalBuildExtension) and ext.store_as_data))

            if self.build_config.package_build_is_in_debug_mode in environ:
                import debugpy
                debugpy.breakpoint()
        
        super().finalize_options()

        with secure_current_dir():
            package_data = self.distribution.package_data or dict()
            
            if self.build_config.package_build_is_in_sdist_mode not in environ:
                for ext in cengal_extensions_store_as_data:
                    files = ext()
                    if files is None:
                        continue

                    if ext.package not in package_data:
                        package_data[ext.package] = list()
                    
                    package_data[ext.package].extend(files)

                from Cython.Build import cythonize
                result_ext_modules = list()
                try:
                    cwd_before_cythonize = getcwd()
                    pyx_flags_dict = prepare_pyx_flags_dict(self.build_config.additional_pyx_flags)
                    result_ext_modules.extend(remove_header_files(
                        cythonize(process_macros(cython_extensions),
                        compiler_directives={'language_level': '3'},
                        compile_time_env = pyx_flags_dict,
                        )))
                except:
                    print(f'DEBUG: {cwd_before_cythonize=} | {getcwd()=}')
                    for extension in cython_extensions:
                        pprint(class_properties_values_including_overrided(extension))
                    
                    raise
                
                result_ext_modules.extend(setuptools_extensions)
                for ext in result_ext_modules:
                    if not hasattr(ext, 'extra_compile_args'):
                        ext.extra_compile_args = list()
                    
                    if ext.extra_compile_args is None:
                        ext.extra_compile_args = list()
                    
                    if 'Darwin' == OS_TYPE:
                        if cpu_info().is_x86:
                            ext.extra_compile_args += ['-arch', 'x86_64']
                        elif cpu_info().is_arm:
                            ext.extra_compile_args += ['-arch', 'arm64']
                
                self.distribution.package_data = package_data
                self.distribution.ext_modules = result_ext_modules

                if self.build_config.package_build_is_in_debug_mode in environ:
                    debugpy.breakpoint()
                    print()



#!/usr/bin/env python
# coding=utf-8

# Copyright © 2012-2024 ButenkoMS. All rights reserved. Contacts: <gtalk@butenkoms.space>
# 
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
# 
#     http://www.apache.org/licenses/LICENSE-2.0
# 
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


__all__ = ['AsyncioLoop', 'AsyncioLoopRequest', 'AsyncioLoopWasNotSetError', 'run_in_thread_pool', 'run_in_thread_pool_fast']

from cengal.parallel_execution.coroutines.coro_scheduler import *
from cengal.parallel_execution.coroutines.coro_tools.await_coro import create_task
from cengal.parallel_execution.coroutines.coro_tools.await_coro import task_in_thread_pool
from cengal.parallel_execution.coroutines.coro_standard_services.loop_yield import gly, agly, CoroPriority, LoopYieldPriorityScheduler
from cengal.parallel_execution.coroutines.coro_standard_services.simple_yield import Yield
from cengal.parallel_execution.coroutines.coro_standard_services.sleep import Sleep
from cengal.parallel_execution.coroutines.coro_standard_services.put_coro import PutCoro
from cengal.parallel_execution.coroutines.coro_standard_services.throw_coro import ThrowCoro
from cengal.parallel_execution.coroutines.coro_standard_services.kill_coro import KillCoro
from cengal.parallel_execution.coroutines.coro_standard_services.wait_coro import WaitCoro, WaitCoroRequest, CoroutineNotFoundError
from cengal.parallel_execution.asyncio.run_loop import run_forever, cancel_all_tasks
from cengal.parallel_execution.asyncio.atasks import create_task_awaitable
from cengal.file_system.file_manager import path_relative_to_current_dir
from cengal.time_management.cpu_clock_cycles import perf_counter
from cengal.time_management.sleep_tools import get_usable_min_sleep_interval, get_min_sleep_interval
from cengal.data_manipulation.serialization import *
from cengal.introspection.inspect import get_exception
from cengal.math.numbers import RationalNumber
from typing import Callable, Tuple, List, Any, Dict, Awaitable, Type
import sys
import os
from asyncio import AbstractEventLoop, get_event_loop, get_running_loop, Task as asyncio_Task, sleep as asyncio_sleep
from asyncio.coroutines import _is_coroutine
from cengal.code_flow_control.args_manager import args_kwargs
from types import coroutine as types_coroutine
from asyncio import coroutines
from asyncio import events
from asyncio import tasks
import threading


"""
Module Docstring
Docstrings: http://www.python.org/dev/peps/pep-0257/
"""

__author__ = "ButenkoMS <gtalk@butenkoms.space>"
__copyright__ = "Copyright © 2012-2024 ButenkoMS. All rights reserved. Contacts: <gtalk@butenkoms.space>"
__credits__ = ["ButenkoMS <gtalk@butenkoms.space>", ]
__license__ = "Apache License, Version 2.0"
__version__ = "4.4.0"
__maintainer__ = "ButenkoMS <gtalk@butenkoms.space>"
__email__ = "gtalk@butenkoms.space"
# __status__ = "Prototype"
__status__ = "Development"
# __status__ = "Production"


class AsyncioLoopWasNotSetError(Exception):
    pass


class ExternalAsyncioLoopAlreadyExistsError(Exception):
    pass


class WaitingCancelled(Exception):
    pass


WAITING_FOR_NEW_REQUESTS_EVENT = 'AsyncioLoopRequest - WaitingForNewRequestsEvent'


class AsyncioLoopRequest(ServiceRequest):
    def inherit_surrounding_loop(self) -> AbstractEventLoop:
        return self._save(0)

    def start_internal_loop(self, main_awaitable: Optional[Awaitable] = None, priority: Optional[CoroPriority] = None, interrupt_when_no_requests: Optional[bool] = None, debug: bool = False) -> AbstractEventLoop:
        return self._save(1, main_awaitable, priority, interrupt_when_no_requests, debug)

    def ensure_loop(self, main_awaitable: Optional[Awaitable] = None, priority: Optional[CoroPriority] = None, interrupt_when_no_requests: Optional[bool] = None) -> AbstractEventLoop:
        return self._save(2, main_awaitable, priority, interrupt_when_no_requests)

    def set(self, async_loop) -> None:
        return self._save(3, async_loop)

    def get(self) -> AbstractEventLoop:
        return self._save(4)

    def wait(self, awaitable: Awaitable) -> Any:
        return self._save(5, awaitable, False, True)

    def create_task(self, awaitable: Awaitable) -> asyncio_Task:
        return self._save(6, awaitable)

    def _internal_loop_yield(self) -> None:
        return self._save(7)

    def turn_on_loops_intercommunication(self, turn_on: bool = True) -> Optional[Callable]:
        return self._save(8, turn_on)

    def _wait_intercommunication(self, awaitable: Awaitable) -> Any:
        return self._save(9, awaitable, True, True)

    def wait_idle(self, awaitable: Awaitable) -> Any:
        return self._save(10, awaitable, False, False)

    def use_higher_level_sleep_manager(self, use_higher_level_sleep_manager: bool = True) -> Any:
        return self._save(12, use_higher_level_sleep_manager)

    def turn_on_low_latency_io_mode(self, low_latency_io_mode: bool = True) -> Any:
        return self._save(13, low_latency_io_mode)


async def asyncio_coro_sleep_0():
    return await asyncio_sleep(0)


def request_giver(request):
    yield request

@types_coroutine
def asyncio_coro_request(request):
    return (yield from request_giver(request))

asyncio_coro_request._is_coroutine = _is_coroutine


class AsyncioLoop(Service, EntityStatsMixin):
    def __init__(self, loop: CoroSchedulerType):
        super(AsyncioLoop, self).__init__(loop)
        self.async_loop: Optional[AbstractEventLoop] = None
        self.internal_async_loop: Optional[AbstractEventLoop] = None
        self._internal_loop_holding_coro: Optional[CoroWrapperBase] = None
        self.internal_loop_start_waiters: Set[CoroID] = set()
        self.need_to_stop_internal_loop: bool = False
        self.internal_loop_creation_error: Optional[Exception] = None
        self.internal_loop_in_yield: bool = False
        self.waiting_for_new_requests: bool = False
        self.loops_intercommunication: bool = False
        self._previous_on_wrong_request = None
        self.intercommunication_requests_coro_ids: Set[CoroID] = set()

        self._request_workers = {
            0: self._on_inherit_surrounding_loop,
            1: self._on_start_internal_loop,
            2: self._on_ensure_loop,
            3: self._on_set,
            4: self._on_get,
            5: self._on_await,
            6: self._on_create_task,
            7: self._on__internal_loop_yield,
            8: self._on_turn_on_loops_intercommunication,
            9: self._on_await,
            10: self._on_await,
            11: self._on__internal_wait_for_new_requests, 
            12: self._on__use_higher_level_sleep_manager, 
            13: self._on__low_latency_io_mode, 
        }
        
        self.pending_requests_num: int = 0
        self.new_requests_num: int = 0
        self.no_idle_calls: Set[CoroID] = set()
        self.results: Dict[CoroID, Tuple[Any, Exception]] = dict()
        self._waiting_coro_id: Optional[CoroID] = None
        self._original_loop_class: Type = None
        self._idle_for: Optional[RationalNumber] = None  # in seconds
        self.use_higher_level_sleep_manager: bool = False
        self.current_on_idle_handler: Optional[Callable] = None
        self.low_latency_io_mode: int = 0
    
    def destroy(self):
        if self.internal_async_loop is not None:
            events._set_running_loop(None)

    def _on_system_loop_idle(self, next_event_after: Optional[RationalNumber]):
        if next_event_after is None:
            self._idle_for = max(0.001, get_usable_min_sleep_interval())
        else:
            self._idle_for = max(0.001, next_event_after)

    def get_entity_stats(self, stats_level: 'EntityStatsMixin.StatsLevel' = EntityStatsMixin.StatsLevel.debug) -> Tuple[str, Dict[str, Any]]:
        return type(self).__name__, {
            'pending requests num': self.pending_requests_num
        }

    def single_task_registration_or_immediate_processing(
            self, request: Optional[AsyncioLoopRequest]=None) -> Tuple[bool, None, None]:
        if request is not None:
            return self.resolve_request(request)
        return True, None, None

    def full_processing_iteration(self):
        from cengal.parallel_execution.coroutines.coro_standard_services.async_event_bus import AsyncEventBusRequest, try_send_async_event
        if (self._waiting_coro_id is not None) and self.new_requests_num:
            # throw_coro_service: ThrowCoro = self._loop.get_service_instance(ThrowCoro)
            # throw_coro_service._add_direct_request(self._waiting_coro_id, WaitingCancelled)
            kill_coro_service: KillCoro = self._loop.get_service_instance(KillCoro)
            kill_coro_service._add_direct_request(self._waiting_coro_id)
            self._waiting_coro_id = None
            try_send_async_event(self._loop, WAITING_FOR_NEW_REQUESTS_EVENT, None)

        if self.internal_loop_in_yield:
            if self.pending_requests_num:
                self.register_response(self._internal_loop_holding_coro.coro_id, None, None)
                self.internal_loop_in_yield = False

        if self.internal_loop_start_waiters:
            loop_response = None
            exception_response = None
            if self.internal_async_loop is not None:
                loop_response = self.async_loop = self.internal_async_loop
            elif self.internal_loop_creation_error is not None:
                exception_response = self.internal_loop_creation_error
            
            if loop_response or exception_response:
                for coro_id in self.internal_loop_start_waiters:
                    self.register_response(coro_id, loop_response, exception_response)

                self.internal_loop_start_waiters = type(self.internal_loop_start_waiters)()

        for coro_id, response in self.results.items():
            result, exception = response
            if coro_id in self.intercommunication_requests_coro_ids:
                self.intercommunication_requests_coro_ids.remove(coro_id)
                self._responses.append(DirectResponse(coro_id, type(self), result, exception))
            else:
                self.register_response(coro_id, result, exception)
        
        self.pending_requests_num -= len(self.results)
        self.results = type(self.results)()
        
        if not self.no_idle_calls:
            self.make_dead()
    
    def is_low_latency(self) -> bool:
        return True

    def in_work(self) -> bool:
        result: bool = bool(self.internal_loop_start_waiters) | bool(self.pending_requests_num) | bool(self.results)
        return self.thrifty_in_work(result)
    
    def _on_inherit_surrounding_loop(self) -> Tuple[bool, Optional[AbstractEventLoop], Exception]:
        exception = None
        try:
            self.async_loop = get_running_loop()
        except:
            exception = get_exception()

        return True, self.async_loop, exception
    
    def _on_start_internal_loop(self, main_awaitable: Optional[Awaitable] = None, priority: Optional[CoroPriority] = None, interrupt_when_no_requests: Optional[bool] = None, debug: bool = False) -> Tuple[bool, Optional[AbstractEventLoop], Exception]:
        if self.internal_async_loop:
            self.async_loop = self.internal_async_loop
            return True, self.async_loop, None
        
        coro_id = self.current_caller_coro_info.coro_id
        self.internal_loop_start_waiters.add(coro_id)
        if self._internal_loop_holding_coro is None:
            self.internal_async_loop = None
            self.internal_loop_creation_error = None
            try:
                if self._is_asyncio_loop_has_run_once_method():
                    internal_loop_holding_coro_worker = ExplicitWorker(CoroType.awaitable, _internal_loop_holding_coro_run_once_based)
                else:
                    internal_loop_holding_coro_worker = ExplicitWorker(CoroType.greenlet, _internal_loop_holding_coro)
            except ExternalAsyncioLoopAlreadyExistsError as ex:
                return True, None, ex
            
            self._internal_loop_holding_coro = self._loop.put_coro(internal_loop_holding_coro_worker, self, main_awaitable, priority, interrupt_when_no_requests, debug)
            if interrupt_when_no_requests:
                self._internal_loop_holding_coro.is_background_coro = True

        return False, None, None
    
    def _is_asyncio_loop_has_run_once_method(self) -> bool:
        result = False
        if events._get_running_loop() is not None:
            raise ExternalAsyncioLoopAlreadyExistsError

        loop = None
        try:
            loop = events.new_event_loop()
            if hasattr(loop, '_run_once'):
                result = True
        finally:
            if loop is not None:
                loop.close()
        
        return result
    
    def _on_ensure_loop(self, main_awaitable: Optional[Awaitable] = None, priority: Optional[CoroPriority] = None, interrupt_when_no_requests: Optional[bool] = None) -> Tuple[bool, Optional[AbstractEventLoop], Exception]:
        result_exists, result, exception = self._on_get()
        if result and (exception is None):
            return result_exists, result, exception

        result_exists, result, exception = self._on_inherit_surrounding_loop()
        if result and (exception is None):
            return result_exists, result, exception
        
        return self._on_start_internal_loop(main_awaitable, priority, interrupt_when_no_requests)

    def _on_set(self, async_loop):
        self.async_loop = async_loop
        return True, None, None

    def _on_get(self):
        if self.async_loop is None:
            exception = AsyncioLoopWasNotSetError()
        else:
            exception = None
        
        return True, self.async_loop, exception
    
    def register_await_response(self, coro_id: CoroID, response: Any, exception: Optional[Exception]):
        self.results[coro_id] = (response, exception)
        self.no_idle_calls.discard(coro_id)
        self.make_live()

    def _on_await(self, awaitable: Awaitable, intercommunication_request: bool = False, prevent_idle: bool = True) -> Tuple[bool, Any, Optional[Exception]]:
        if self.async_loop is None:
            return True, None, AsyncioLoopWasNotSetError()
        
        coro_id = self.current_caller_coro_info.coro_id
        if intercommunication_request:
            self.intercommunication_requests_coro_ids.add(coro_id)
        
        async def awaiting_worker(asyncio_loop_instance: AsyncioLoop, coro_id: CoroID, awaitable: Awaitable):
            exception = None
            result = None
            try:
                result = await awaitable
            except:
                exception = get_exception()
            
            asyncio_loop_instance.register_await_response(coro_id, result, exception)
        
        create_task(self.async_loop, awaiting_worker, self, coro_id, awaitable)
        
        self.pending_requests_num += 1
        if prevent_idle:
            self.no_idle_calls.add(coro_id)
        
        self.make_live()
        
        return False, None, None

    def _on_create_task(self, awaitable: Awaitable) -> Tuple[bool, Optional[asyncio_Task], Optional[Exception]]:
        if self.async_loop is None:
            return True, None, AsyncioLoopWasNotSetError()

        async def awaiting_wrapper(awaitable: Awaitable):
            return await awaitable

        task: asyncio_Task = create_task(self.async_loop, awaiting_wrapper, awaitable)
        return True, task, None
    
    def _on__internal_loop_yield(self) -> Tuple[bool, None, None]:
        if self.pending_requests_num:
            return True, None, None
        else:
            self.internal_loop_in_yield = True
            return False, None, None
    
    def _on__internal_wait_for_new_requests(self) -> Tuple[bool, None, None]:
        if self.new_requests_num:
            return True, None, None
        else:
            self.waiting_for_new_requests = True
            return False, None, None
    
    def register_new_asyncio_request(self) -> None:
        self.new_requests_num += 1
        self.make_live()
    
    def add_on_idle_handler(self) -> None:
        if not self.use_higher_level_sleep_manager:
            self.current_on_idle_handler = self._on_system_loop_idle
            self._loop.on_idle_handlers.add(self._on_system_loop_idle)
    
    def discard_on_idle_handler(self) -> None:
        self._loop.on_idle_handlers.discard(self._on_system_loop_idle)
        self.current_on_idle_handler = None
    
    def _on__use_higher_level_sleep_manager(self, use_higher_level_sleep_manager: bool) -> Tuple[bool, Optional[None], None]:
        self.use_higher_level_sleep_manager = use_higher_level_sleep_manager
        if use_higher_level_sleep_manager:
            if self.current_on_idle_handler is not None:
                self._loop.on_idle_handlers.add(self.current_on_idle_handler)
        else:
            self.discard_on_idle_handler()
        
        return True, None, None

    def _on__low_latency_io_mode(self, low_latency_io_mode: bool) -> Tuple[bool, Optional[bool], None]:
        buff_low_latency_io_mode = self.low_latency_io_mode > 0
        if low_latency_io_mode:
            self.low_latency_io_mode += 1
        else:
            self.low_latency_io_mode -= 1
        
        return True, buff_low_latency_io_mode, None
    
    def _on_turn_on_loops_intercommunication(self, turn_on: bool) -> Tuple[bool, Optional[Callable], None]:
        result = self._loop.on_wrong_request
        if turn_on:
            self._previous_on_wrong_request = self._loop.on_wrong_request
            self.loops_intercommunication = True
            self._loop.on_wrong_request = self._on_wrong_request
        else:
            if self.loops_intercommunication:
                self._loop.on_wrong_request = self._previous_on_wrong_request
                self.loops_intercommunication = False
            
            self._previous_on_wrong_request = None
        
        return True, result, None
    
    def _on_wrong_request(self, coro: CoroWrapperBase, request: Any) -> Request:
        if request is None:
            args, kwargs = args_kwargs(AsyncioLoopRequest()._wait_intercommunication(asyncio_coro_sleep_0()))
            result: Request = Request(coro, type(self), *args, **kwargs)
        else:
            args, kwargs = args_kwargs(AsyncioLoopRequest()._wait_intercommunication(asyncio_coro_request(request)))
            result = Request(coro, type(self), *args, **kwargs)

        return result
    
    def inline_get(self):
        if self.async_loop is None:
            raise AsyncioLoopWasNotSetError
        else:
            return self.async_loop
    
    def inline_set_internal_loop(self, loop, exception: Optional[Exception]):
        self.internal_async_loop = loop
        self.internal_loop_creation_error = exception
        self.make_live()
    
    def is_need_to_yield_internal_loop(self) -> bool:
        return not self.pending_requests_num


AsyncioLoopRequest.default_service_type = AsyncioLoop


def _internal_loop_holding_coro(i: Interface, service: AsyncioLoop, main_awaitable: Optional[Awaitable] = None, priority: Optional[CoroPriority] = None, interrupt_when_no_requests: Optional[bool] = None, debug: bool = False):
    ly = None
    if priority is not None:
        ly = gly(priority)

    async def main_wrapper(service: AsyncioLoop, main_awaitable: Optional[Awaitable] = None, priority: Optional[CoroPriority] = None, interrupt_when_no_requests: Optional[bool] = None):
        loop: AbstractEventLoop = get_event_loop()
        service.inline_set_internal_loop(loop, None)
        if interrupt_when_no_requests is None:
            interrupt_when_no_requests = main_awaitable is None
        
        def on_loop_simple_yield():
            if interrupt_when_no_requests and service.is_need_to_yield_internal_loop():
                i(AsyncioLoop, AsyncioLoopRequest()._internal_loop_yield())
            else:
                i(Yield)
            
            if not service.need_to_stop_internal_loop:
                if service._idle_for is None:
                    loop.call_soon(on_loop_simple_yield)
                else:
                    idle_for = service._idle_for
                    service._idle_for = None
                    loop.call_later(idle_for, on_loop_simple_yield)
        
        def on_loop_loop_yield():
            if interrupt_when_no_requests and service.is_need_to_yield_internal_loop():
                i(AsyncioLoop, AsyncioLoopRequest()._internal_loop_yield())
            else:
                ly()
            
            if not service.need_to_stop_internal_loop:
                if service._idle_for is None:
                    loop.call_soon(on_loop_loop_yield)
                else:
                    idle_for = service._idle_for
                    service._idle_for = None
                    loop.call_later(idle_for, on_loop_loop_yield)
        
        if ly is None:
            loop.call_soon(on_loop_simple_yield)
        else:
            loop.call_soon(on_loop_loop_yield)

        if main_awaitable is not None:
            create_task_awaitable(main_awaitable)
    
    exception = None
    try:
        service.add_on_idle_handler()
        run_forever(main_wrapper(service, main_awaitable, priority, interrupt_when_no_requests), debug=debug)
    except:
        exception = get_exception()
        # TODO: do something with exception
    finally:
        try:
            service._loop.on_idle_handlers.discard(service._on_system_loop_idle)
            service.current_on_idle_handler = None
        except ValueError:
            pass

        service.inline_set_internal_loop(None, exception)


async def _internal_loop_holding_coro_run_once_based(i: Interface, service: AsyncioLoop, main_awaitable: Optional[Awaitable] = None, priority: Optional[CoroPriority] = None, interrupt_when_no_requests: Optional[bool] = None, debug: bool = False):
    from cengal.parallel_execution.coroutines.coro_standard_services.async_event_bus import AsyncEventBusRequest, try_send_async_event
    from .known_asyncio_compatible_loops import prepare_loop, restore_loop

    cs: CoroSchedulerType = current_coro_scheduler()
    lyps: LoopYieldPriorityScheduler = cs.get_service_instance(LoopYieldPriorityScheduler)
    umsi: RationalNumber = get_usable_min_sleep_interval()

    ly = None
    if priority is not None:
        ly = await agly(priority)

    async def main_wrapper_for_run_once(service: AsyncioLoop, main_awaitable: Optional[Awaitable] = None):
        loop: AbstractEventLoop = get_event_loop()
        service.inline_set_internal_loop(loop, None)
        if main_awaitable is not None:
            create_task_awaitable(main_awaitable)
    
    if interrupt_when_no_requests is None:
        interrupt_when_no_requests = main_awaitable is None
    
    main_wrapper_for_run_once_coro = main_wrapper_for_run_once(service, main_awaitable)
    
    exception = None
    try:
        if events._get_running_loop() is not None:
            raise RuntimeError(
                'run_forever() cannot be called from a running event loop')

        if not coroutines.iscoroutine(main_wrapper_for_run_once_coro):
            raise ValueError('a coroutine was expected, got {!r}'.format(main_wrapper_for_run_once))

        loop = events.new_event_loop()
        service._original_loop_class = prepare_loop(loop)
        try:
            events.set_event_loop(loop)
            loop.set_debug(debug)
            loop.create_task(main_wrapper_for_run_once_coro)
            loop._check_closed()
            loop._check_running()
            loop._set_coroutine_origin_tracking(loop._debug)
            loop._thread_id = threading.get_ident()

            old_agen_hooks = sys.get_asyncgen_hooks()
            sys.set_asyncgen_hooks(firstiter=loop._asyncgen_firstiter_hook,
                                finalizer=loop._asyncgen_finalizer_hook)
            try:
                events._set_running_loop(loop)
                while True:
                    if ly is None:
                        await i(Yield)
                    else:
                        await ly()

                    loop.call_soon(lambda: None)
                    # if (not loop._ready) and (not loop._stopping) and (not loop._scheduled):
                    #     loop.call_soon(lambda: None)
                    
                    loop._run_once()
                    if loop._stopping:
                        break

                    if (not loop._ready) and (not loop._stopping):
                        if not loop._scheduled:
                            await i(AsyncioLoop, AsyncioLoopRequest()._internal_loop_yield())
                            continue
                        else:
                            when = loop._scheduled[0]._when
                            from asyncio.base_events import MAXIMUM_SELECT_TIMEOUT
                            timeout = min(max(0, when - loop.time()), MAXIMUM_SELECT_TIMEOUT)
                            # if get_min_sleep_interval() <= timeout:
                            #     usable_min_sleep_interval = get_usable_min_sleep_interval()
                            #     if usable_min_sleep_interval > timeout:
                            #         timeout = usable_min_sleep_interval

                            if (not service.low_latency_io_mode) and (0.001 <= timeout) and (not service.results):
                                service.new_requests_num = 0
                                service.make_dead()
                                async def waiting_coro(i: Interface, timeout: float):
                                    from cengal.parallel_execution.coroutines.coro_standard_services.async_event_bus import AsyncEventBusRequest, try_send_async_event
                                    try:
                                        await i(Sleep, timeout)
                                        await i(AsyncEventBusRequest().send_event(WAITING_FOR_NEW_REQUESTS_EVENT, None))
                                    except WaitingCancelled:
                                        print(f'AsyncIoLoop {datetime.now().strftime("%H:%M:%S.%f")} >> WaitingCancelled')
                                
                                from datetime import datetime
                                try:
                                    # TODO: reimplement it in more efficient and elegant way. Btw: Sleep currently will not cancel an event upon coro destroyed
                                    lyps_max_delay = lyps.max_delay
                                    new_max_timeout = umsi if lyps_max_delay < umsi else lyps_max_delay
                                    new_timeout = timeout if new_max_timeout > timeout else new_max_timeout
                                    waiting_coro_id = await i(PutCoro, waiting_coro, new_timeout)
                                    service._waiting_coro_id = waiting_coro_id
                                    # await i(WaitCoro, WaitCoroRequest().single(waiting_coro_id))
                                    await i(AsyncEventBusRequest().wait(WAITING_FOR_NEW_REQUESTS_EVENT))
                                    # print(f'AsyncIoLoop {datetime.now().strftime("%H:%M:%S.%f")} >> WAITING_FOR_NEW_REQUESTS_EVENT')
                                except (WaitingCancelled, CoroutineNotFoundError):
                                    print(f'AsyncIoLoop {datetime.now().strftime("%H:%M:%S.%f")} >> WaitingCancelled or CoroutineNotFoundError')
                                    pass
                                except:
                                    print(f'AsyncIoLoop {datetime.now().strftime("%H:%M:%S.%f")} >> {get_exception()}')
                                    raise

                    if interrupt_when_no_requests and service.is_need_to_yield_internal_loop():
                        await i(AsyncioLoop, AsyncioLoopRequest()._internal_loop_yield())
                        continue
            finally:
                loop._stopping = False
                loop._thread_id = None
                events._set_running_loop(None)
                loop._set_coroutine_origin_tracking(False)
                sys.set_asyncgen_hooks(*old_agen_hooks)
        finally:
            restore_loop(loop, service._original_loop_class)
            try:
                cancel_all_tasks(loop)
                loop.run_until_complete(loop.shutdown_asyncgens())
            finally:
                events.set_event_loop(None)
                loop.close()
    except:
        exception = get_exception()
    finally:
        service.inline_set_internal_loop(None, exception)


def run_in_thread_pool_fast(interface: Interface, function: Callable, *args, **kwargs) -> Any:
    async def task_wrapper(loop, *args, **kwargs):
        return await task_in_thread_pool(loop, function, *args, **kwargs)

    return interface(AsyncioLoop, AsyncioLoopRequest().wait(task_wrapper(
        interface._loop.get_service_instance(AsyncioLoop).inline_get(), 
        *args, **kwargs)))


def run_in_thread_pool(function: Callable, *args, **kwargs) -> Any:
    return run_in_thread_pool_fast(current_interface(), function, *args, **kwargs)



#!/usr/bin/env python
# coding=utf-8

# Copyright © 2012-2024 ButenkoMS. All rights reserved. Contacts: <gtalk@butenkoms.space>
# 
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
# 
#     http://www.apache.org/licenses/LICENSE-2.0
# 
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from .multiprocessing_task_runner import SubprocessWorker
from cengal.data_generation.id_generator import IDGenerator

"""
Module Docstring
Docstrings: http://www.python.org/dev/peps/pep-0257/
"""

__author__ = "ButenkoMS <gtalk@butenkoms.space>"
__copyright__ = "Copyright © 2012-2024 ButenkoMS. All rights reserved. Contacts: <gtalk@butenkoms.space>"
__credits__ = ["ButenkoMS <gtalk@butenkoms.space>", ]
__license__ = "Apache License, Version 2.0"
__version__ = "4.4.0"
__maintainer__ = "ButenkoMS <gtalk@butenkoms.space>"
__email__ = "gtalk@butenkoms.space"
# __status__ = "Prototype"
__status__ = "Development"
# __status__ = "Production"


class SingleTaskInfo:
    def __init__(self, process=None, taskId=None):
        super().__init__()
        self.process = process
        self.isNeedToBeClosed = False
        self.uncompletedTasksQnt = 0  # хочу что бы оно было тут, а не в классе процесса: хочу чтобы класс процесса был
            # thread safe без локов
        self.taskId = taskId


class TaskPool:
    def __init__(self, processesQnt, working_function, initiation_function=None):
        super().__init__()
        self._processesQnt = processesQnt
        self._working_function = working_function
        self._initiation_function = initiation_function
        self._processesIdGenerator = IDGenerator()
        self._taskList = set()
        self._taskDict = dict()
        for processNumber in range(self._processesQnt):
            taskId = self._processesIdGenerator()
            process = SubprocessWorker(self._working_function, self._initiation_function)
            task = SingleTaskInfo(process, taskId)
            self._taskList.add(taskId)
            self._taskDict[taskId] = task
        ...

    def get_processes_qnt(self):
        return self._processesQnt

    def set_processes_qnt(self, processesQnt):
        ...

    def _send_data_to_subprocess(self, inputData, processId):
        ...

    def _get_answer_from_subprocess(self, processId):
        ...

    def send_data_to_the_pool(self, inputData):
        ...

    def get_answer_from_the_pool(self):
        ...

    def send_list_of_data_to_the_pool(self, listOfData):
        ...

    def get_list_of_answers_from_the_pool(self):
        ...

    def stop_pool(self):
        ...

    ...



def rpc_blocking():
    pass



#!/usr/bin/env python
# coding=utf-8

# Copyright © 2012-2024 ButenkoMS. All rights reserved. Contacts: <gtalk@butenkoms.space>
# 
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
# 
#     http://www.apache.org/licenses/LICENSE-2.0
# 
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import errno
from contextlib import contextmanager
import shutil
import os
import os.path
from pathlib import PurePath
from enum import Enum
import hashlib
from cengal.introspection.inspect import frame
from typing import Optional, Callable, Set, Tuple, Union, List, Any

"""
Module Docstring
Docstrings: http://www.python.org/dev/peps/pep-0257/
"""

__author__ = "ButenkoMS <gtalk@butenkoms.space>"
__copyright__ = "Copyright © 2012-2024 ButenkoMS. All rights reserved. Contacts: <gtalk@butenkoms.space>"
__credits__ = ["ButenkoMS <gtalk@butenkoms.space>", ]
__license__ = "Apache License, Version 2.0"
__version__ = "4.4.0"
__maintainer__ = "ButenkoMS <gtalk@butenkoms.space>"
__email__ = "gtalk@butenkoms.space"
# __status__ = "Prototype"
__status__ = "Development"
# __status__ = "Production"


class FilteringType(Enum):
    including = 0  # include only files with extensions within the set of extensions
    excluding = 1  # filter out files with extensions within the set of extensions
    off = 2  # any file will fit criteria


def filtered_file_list(root_dir: str, filtering_type: FilteringType, extentions_set: Optional[Set] = None) -> Tuple:
    """_summary_

    Args:
        root_dir (str): r'C:\dir\path'
        filtering_type (FilteringType): _description_
        extentions_set (Optional[Set], optional): {'.upk', '.txt'}. Defaults to None.

    Raises:
        Exception: _description_

    Returns:
        Tuple: (dirpath, dirnames, new_filenames)
    """

    if (FilteringType.off != filtering_type) and (extentions_set is None):
        raise Exception('extentions_set can\'t be None with this filtering type')

    result = None
    raw_result = next(os.walk(root_dir))
    dirpath = raw_result[0]
    dirnames = raw_result[1]
    filenames = raw_result[2]

    new_filenames = list()
    for filename in filenames:
        if FilteringType.off == filtering_type:
            new_filenames.append(filename)
        elif FilteringType.including == filtering_type:
            extention = os.path.splitext(filename)[1]
            if extention in extentions_set:
                new_filenames.append(filename)
        elif FilteringType.excluding == filtering_type:
            extention = os.path.splitext(filename)[1]
            if extention not in extentions_set:
                new_filenames.append(filename)

    result = (dirpath, dirnames, new_filenames)
    return result


def filtered_file_list_traversal(root_dir: str, filtering_type: FilteringType, extentions_set: Optional[Set] = None, 
                                 remove_empty_items: bool =False, use_spinner: bool = False):
    """
    :param root_dir: str(); r'C:\dir\path'
    :param filtering_type: FilteringType()
    :param extentions_set: set(); {'.upk', '.txt'}
    :return: list() of tuple(); list of (dirpath, dirnames, new_filenames)
    """

    if (FilteringType.off != filtering_type) and (extentions_set is None):
        raise Exception('extentions_set can\'t be None with this filtering type')

    spinner = None
    spinner_index = 0
    if use_spinner:
        from progress.spinner import Spinner
        spinner = Spinner('Loading ')
    result_list = list()
    for raw_result in os.walk(root_dir):
        dirpath = raw_result[0]
        dirnames = raw_result[1]
        filenames = raw_result[2]

        new_filenames = list()
        for filename in filenames:
            if FilteringType.off == filtering_type:
                new_filenames.append(filename)
            elif FilteringType.including == filtering_type:
                extention = os.path.splitext(filename)[1]
                if extention in extentions_set:
                    new_filenames.append(filename)
            elif FilteringType.excluding == filtering_type:
                extention = os.path.splitext(filename)[1]
                if extention not in extentions_set:
                    new_filenames.append(filename)
        if remove_empty_items:
            if len(new_filenames) > 0:
                result = (dirpath, dirnames, new_filenames)
                result_list.append(result)
        else:
            result = (dirpath, dirnames, new_filenames)
            result_list.append(result)

        if use_spinner:
            spinner_index += 1
            if spinner_index >= 1000:
                spinner_index = 0
                spinner.next()
    if use_spinner:
        print()

    return result_list


class FilteringEntity(Enum):
    dirpath = 0
    dirname = 1
    filename = 2
    aggregated = 3


def file_list_traversal(root_dir: str, filter: Callable = None, remove_empty_dirpaths=False) -> Union[Any, List[Tuple[str, List[str], List[str]]]]:
    """
    :param root_dir: str(); r'C:\dir\path'
    :param filtering_type: FilteringType()
    :param extentions_set: set(); {'.upk', '.txt'}
    :return: list() of tuple(); list of (dirpath, dirnames, new_filenames)
    """

    if filter is None:
        def default_filter(filtering_entity: FilteringEntity, data):
            if FilteringEntity.aggregated == filtering_entity:
                return data
            else:
                return True

        filter = default_filter

    result_list = list()
    for raw_result in os.walk(root_dir):
        dirpath = raw_result[0]
        dirnames = raw_result[1]
        filenames = raw_result[2]

        if not filter(FilteringEntity.dirpath, (dirpath, dirnames, filenames)):
            continue

        new_dirnames = list()
        for dirname in dirnames:
            if filter(FilteringEntity.dirname, (dirpath, dirname)):
                new_dirnames.append(dirname)

        new_filenames = list()
        for filename in filenames:
            if filter(FilteringEntity.filename, (dirpath, filename)):
                new_filenames.append(filename)
        
        if remove_empty_dirpaths:
            if not new_filenames and not new_dirnames:
                continue

        result = (dirpath, new_dirnames, new_filenames)
        result_list.append(result)

    return filter(FilteringEntity.aggregated, result_list)


def file_list_traversal_ex(
        root_dir: str, 
        filter: Callable = None, 
        remove_empty_dirpaths=False,
        subignore_by_dirname=True,
        subignore_by_dirpath=True,
        subignore: Optional[Set[str]] = None
        ) -> Union[Any, List[Tuple[str, List[str], List[str]]]]:
    """
    :param root_dir: str(); r'C:\dir\path'
    :param filtering_type: FilteringType()
    :param extentions_set: set(); {'.upk', '.txt'}
    :return: list() of tuple(); list of (dirpath, dirnames, new_filenames)
    """

    if filter is None:
        def default_filter(filtering_entity: FilteringEntity, data):
            if FilteringEntity.aggregated == filtering_entity:
                return data
            else:
                return True

        filter = default_filter

    subignore = set() if subignore is None else subignore
    def is_subignored(dir_path: str) -> bool:
        if not subignore:
            return False

        if dir_path in subignore:
            return True
        
        while True:
            new_dir_name = os.path.dirname(dir_path)
            if new_dir_name == dir_path:
                break
            
            if new_dir_name in subignore:
                return True

            dir_path = new_dir_name

        return False

    result_list = list()
    for raw_result in os.walk(root_dir):
        dirpath = raw_result[0]
        dirnames = raw_result[1]
        filenames = raw_result[2]

        if is_subignored(dirpath):
            continue
        
        if not filter(FilteringEntity.dirpath, (dirpath, dirnames, filenames)):
            if subignore_by_dirpath:
                subignore.add(dirpath)
            
            continue

        new_dirnames = list()
        for dirname in dirnames:
            if is_subignored(os.path.join(dirpath, dirname)):
                continue

            if filter(FilteringEntity.dirname, (dirpath, dirname)):
                new_dirnames.append(dirname)
            else:
                if subignore_by_dirname:
                    subignore.add(os.path.join(dirpath, dirname))

        new_filenames = list()
        for filename in filenames:
            if filter(FilteringEntity.filename, (dirpath, filename)):
                new_filenames.append(filename)
        
        if remove_empty_dirpaths:
            if not new_filenames and not new_dirnames:
                continue

        result = (dirpath, new_dirnames, new_filenames)
        result_list.append(result)

    return filter(FilteringEntity.aggregated, result_list)


def clear_dir(full_dir_path):
    full_dir_path = os.path.normpath(full_dir_path)
    if (os.path.exists(full_dir_path) and os.path.isdir(full_dir_path)):
        dir_path, sub_dir_names, file_names = filtered_file_list(full_dir_path, FilteringType.off)
        for sub_dir in sub_dir_names:
            full_sub_dir_path = os.path.join(dir_path, sub_dir)
            shutil.rmtree(full_sub_dir_path, ignore_errors=True)
        for file in file_names:
            full_file_name = os.path.join(dir_path, file)
            os.remove(full_file_name)


@contextmanager
def change_current_dir(new_current_dir):
    new_current_dir = os.path.normpath(new_current_dir)
    cur_dir = os.getcwd()
    os.chdir(new_current_dir)
    try:
        yield
    except:
        raise
    finally:
        os.chdir(cur_dir)


@contextmanager
def secure_current_dir():
    cur_dir = os.getcwd()
    try:
        yield
    except:
        raise
    finally:
        os.chdir(cur_dir)


def get_dir_hash(full_dir_path, hash_format_string=None):
    hash_format_string = hash_format_string or '{} {}'
    result_list = filtered_file_list_traversal(full_dir_path, FilteringType.off)

    dir_content = list()
    for dirpath, dirnames, new_filenames in result_list:
        for filename in new_filenames:
            full_file_name = os.path.join(dirpath, filename)
            with open(full_file_name, 'rb') as file:
                file_content = file.read()
                dir_content.append(file_content)
    dir_content = b''.join(dir_content)
    dir_hash = hash_format_string.format(hashlib.sha512(dir_content).hexdigest(), hex(len(dir_content))[2:])
    return dir_hash


def current_src_dir(depth: Optional[int] = 1):
    """

    :param depth: 0 - path of this file, 1 - path of the caller's file, etc.
    :return:
    """
    depth = depth or 0
    depth += 1
    return os.path.dirname(os.path.realpath(frame(depth).f_code.co_filename))


def ensure_dir(dir_path):
    dir_path = os.path.normpath(dir_path)
    if not (os.path.exists(dir_path) and os.path.isdir(dir_path)):
        try:
            os.makedirs(dir_path, exist_ok=True)
        except OSError as er:
            # Windows can raise spurious ENOTEMPTY errors. See https://github.com/pypa/pip/issues/6426
            if (errno.EEXIST != er.errno) and (errno.ENOTEMPTY != er.errno):
                raise


def ensure_empty_dir(dir_path):
    ensure_dir(dir_path)
    clear_dir(dir_path)


def dir_exists(dir_path) -> bool:
    dir_path: str = os.path.normpath(dir_path)
    return os.path.exists(dir_path) and os.path.isdir(dir_path)


class DirNotExistsError(Exception):
    pass


def dir_empty(dir_path) -> bool:
    dir_path: str = os.path.normpath(dir_path)
    if not dir_exists(dir_path):
        raise DirNotExistsError

    dir_path, sub_dir_names, file_names = filtered_file_list(dir_path, FilteringType.off)
    return (not sub_dir_names) and (not file_names)


def path_relative_to_current_dir(relative_path: Optional[str]=None) -> str:
    relative_path = relative_path or str()
    return os.path.normpath(os.path.join(os.getcwd(), os.path.normpath(relative_path)))



from typing import Dict, List, Any


NimObj = Dict[str, Any]


def serve(addr_list: List[NimObj]) -> List[NimObj]:
    ...


def client() -> NimObj:
    ...


def connect(addr_list: List[NimObj]) -> List[NimObj]:
    ...


def server_disconnect(reactorId: int, connectionId: int) -> bool:
    ...


def client_disconnect(reactorId: int, connectionId: int) -> bool:
    ...


def punch_through(address: NimObj):
    ...


def tick(server_send: Dict[int, List[NimObj]], client_send: Dict[int, List[NimObj]]) -> NimObj:
    ...



#!/usr/bin/env python
# coding=utf-8

# Copyright © 2012-2024 ButenkoMS. All rights reserved. Contacts: <gtalk@butenkoms.space>
# 
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
# 
#     http://www.apache.org/licenses/LICENSE-2.0
# 
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


__all__ = ['WrongTextType', 'KNOWN_BOMS', 'determine_text_bom', 'remove_bom', 'determine_bom_encoding', 'decode_text_and_remove_all_wrong_symbols']


"""
Module Docstring
Docstrings: http://www.python.org/dev/peps/pep-0257/
"""


__author__ = "ButenkoMS <gtalk@butenkoms.space>"
__copyright__ = "Copyright © 2012-2024 ButenkoMS. All rights reserved. Contacts: <gtalk@butenkoms.space>"
__credits__ = ["ButenkoMS <gtalk@butenkoms.space>", ]
__license__ = "Apache License, Version 2.0"
__version__ = "4.4.0"
__maintainer__ = "ButenkoMS <gtalk@butenkoms.space>"
__email__ = "gtalk@butenkoms.space"
# __status__ = "Prototype"
__status__ = "Development"
# __status__ = "Production"


import platform, sys
import codecs
from typing import Optional, Union
from cengal.text_processing.text_processing import Text, DEFAULT_ENCODING, normalize_text, removeprefix


class WrongTextType(Exception):
    pass


KNOWN_BOMS = {
    codecs.BOM_UTF8: 'utf-8',
    codecs.BOM_UTF16_BE: 'utf-16be',
    codecs.BOM_UTF16_LE: 'utf-16le',
    codecs.BOM_UTF32_BE: 'utf-32be',
    codecs.BOM_UTF32_LE: 'utf-32le',
}
KNOWN_BOMS_ORDER = [
    codecs.BOM_UTF8,
    codecs.BOM_UTF32_BE,
    codecs.BOM_UTF32_LE,
    codecs.BOM_UTF16_BE,
    codecs.BOM_UTF16_LE,
]


def determine_text_bom(text: Union[bytes, bytearray]) -> Union[bytes, bytearray]:
    if (not isinstance(text, bytes)) and (not isinstance(text, bytearray)):
        raise WrongTextType
    
    bom_list = list()
    
    absent_bom = b''
    
    if isinstance(text, bytearray):
        for bom in KNOWN_BOMS_ORDER:
            bom_list.append(normalize_text(bom, bytearray))
        
        absent_bom = bytearray(absent_bom)
    else:
        bom_list = list(KNOWN_BOMS_ORDER)

    for bom in bom_list:
        if text.startswith(bom):
            return bom
    
    return absent_bom


def remove_bom(text: Union[bytes, bytearray], bom: Union[bytes, bytearray]) -> Union[bytes, bytearray]:
    return removeprefix(text, bom)


def determine_bom_encoding(bom: Union[bytes, bytearray]) -> Optional[str]:
    bom = normalize_text(bom, bytes)
    return KNOWN_BOMS.get(bom, None)


def decode_text_and_remove_all_wrong_symbols(text: Union[bytes, bytearray], encoding: str) -> str:
    return text.decode(encoding, 'replace')

